"""Functions that can be used to test binary search operators."""
from math import isqrt
from typing import Any, Callable

from numpy.random import Generator, default_rng

from moptipy.api.operators import Op2, check_op2
from moptipy.api.space import Space
from moptipy.tests.component import validate_component
from moptipy.utils.types import type_error


def validate_op2(op2: Op2,
                 search_space: Space = None,
                 make_search_space_element_valid:
                 Callable[[Generator, Any], Any] | None = lambda _, x: x,
                 number_of_samples: int = 100,
                 min_unique_samples: int | Callable[[int, Space], int]
                 = lambda samples, space:
                 max(1, min(samples // 2, isqrt(space.n_points())))) -> None:
    """
    Check whether an object is valid a moptipy binary operator.

    :param op2: the operator
    :param search_space: the search space
    :param make_search_space_element_valid: make a point in the search
        space valid
    :param number_of_samples: the number of times to invoke the operator
    :param min_unique_samples: a lambda for computing the number
    :raises ValueError: if `op2` is not a valid instance of
        :class:`~moptipy.api.operators.Op2`
    :raises TypeError: if incorrect types are encountered
    """
    if not isinstance(op2, Op2):
        raise type_error(op2, "op2", Op2)
    if op2.__class__ == Op2:
        raise ValueError("Cannot use abstract base Op2 directly.")
    check_op2(op2)
    validate_component(op2)

    count: int = 0
    if search_space is not None:
        count += 1
    if make_search_space_element_valid is not None:
        count += 1
    if count <= 0:
        return
    if count < 2:
        raise ValueError(
            "either provide both of search_space and "
            "make_search_space_element_valid or none.")
    if not isinstance(number_of_samples, int):
        raise type_error(number_of_samples, "number_of_samples", int)
    if not (1 <= number_of_samples <= 1_000_000):
        raise ValueError("number_of_samples must be in 1..1_000_000, "
                         f"but is {number_of_samples}.")

    random = default_rng()

    x1 = search_space.create()
    if x1 is None:
        raise ValueError("Space must not return None.")
    x1 = make_search_space_element_valid(random, x1)
    if x1 is None:
        raise ValueError("validator turned point to None?")
    search_space.validate(x1)

    seen = set()

    strstr = search_space.to_str(x1)
    if (not isinstance(strstr, str)) or (len(strstr) <= 0):
        raise ValueError("to_str produces either no string or "
                         f"empty string, namely {strstr}.")
    seen.add(strstr)
    x2: Any = None
    for _ in range(1000):
        x2 = search_space.create()
        if x2 is None:
            raise ValueError("Space must not return None.")
        x2 = make_search_space_element_valid(random, x2)
        if x2 is None:
            raise ValueError("validator turned point to None?")
        if x1 is x2:
            raise ValueError(
                "Search space.create must not return same object instance.")
        search_space.validate(x2)
        strstr = search_space.to_str(x2)
        if (not isinstance(strstr, str)) or (len(strstr) <= 0):
            raise ValueError("to_str produces either no string or "
                             f"empty string, namely {strstr}.")
        seen.add(strstr)
        if len(seen) > 1:
            break
    if len(seen) <= 1:
        raise ValueError("failed to create two different initial elements.")

    x3 = search_space.create()
    if x3 is None:
        raise ValueError("Space must not return None.")
    if (x1 is x3) or (x2 is x3):
        raise ValueError(
            "Search space.create must not return same object instance.")

    if not (hasattr(op2, "op2") and callable(getattr(op2, "op2"))):
        raise ValueError("op2 must have method op2.")
    for _ in range(number_of_samples):
        op2.op2(random, x3, x2, x1)
        search_space.validate(x3)
        strstr = search_space.to_str(x3)
        if (not isinstance(strstr, str)) or (len(strstr) <= 0):
            raise ValueError("to_str produces either no string or "
                             f"empty string, namely '{strstr}'.")
        seen.add(strstr)

    expected: int
    if callable(min_unique_samples):
        expected = min_unique_samples(number_of_samples, search_space)
    else:
        expected = min_unique_samples
    if not isinstance(expected, int):
        raise type_error(expected, "expected", int)
    if expected > number_of_samples:
        raise ValueError(
            f"number of expected unique samples {expected} cannot be larger "
            f"than total number of samples {number_of_samples}.")
    if expected <= 0:
        raise ValueError(f"expected number of unique values must be positive,"
                         f" but is {expected}.")
    if len(seen) < expected:
        raise ValueError(
            f"It is expected that at least {expected} different elements "
            "will be created by unary search operator from "
            f"{number_of_samples} samples, but we only "
            f"got {len(seen)} different points.")



"""
moptipy is the *Metaheuristic Optimization in Python* Package.

This is the basic package of our moptipy project.
It provides several sub-packages:

* :py:mod:`moptipy.api` the API for implementing optimization
  algorithms and -problems
* :py:mod:`moptipy.spaces` several pre-defined search spaces
* :py:mod:`moptipy.utils` utility classes for, e.g., logging
  and naming
"""
from typing import Final

import moptipy.version

__version__: Final[str] = moptipy.version.__version__



"""A class for representing Gantt charts as objects."""
from typing import Final

import numpy as np

from moptipy.api.logging import SECTION_RESULT_Y, SECTION_SETUP
from moptipy.evaluation.log_parser import LogParser
from moptipy.examples.jssp.instance import Instance
from moptipy.utils.types import type_error


# start book
class Gantt(np.ndarray):
    """
    A class representing Gantt charts.

    A Gantt chart is a diagram that visualizes when a job on a given
    machine begins or ends. We here represent it as a three-dimensional
    matrix. This matrix has one row for each machine and one column for
    each operation on the machine.
    In each cell, it holds three values: the job ID, the start, and the
    end time of the job on the machine. The Gantt chart has the
    additional attribute `instance` which references the JSSP instance
    for which the chart is constructed.
    Gantt charts must only be created by an instance of
    :class:`moptipy.examples.jssp.gantt_space.GanttSpace`.
    """

    #: the JSSP instance for which the Gantt chart is created
    instance: Instance
# end book

    def __new__(cls, space) -> "Gantt":
        """
        Create the Gantt chart.

        :param space: the Gantt space for which the instance is created.
        """
        gnt: Final[Gantt] = np.ndarray.__new__(Gantt, space.shape, space.dtype)
        #: the JSSP instance for which the Gantt chart is created
        gnt.instance = space.instance
        return gnt

    @staticmethod
    def from_log(file: str,
                 instance: Instance | None = None) -> "Gantt":
        """
        Load a Gantt chart from a log file.

        :param file: the log file path
        :param instance: the optional JSSP instance: if `None` is provided,
            we try to load it from the resources
        :returns: the Gantt chart
        """
        parser: Final[_GanttParser] = _GanttParser(instance)
        parser.parse_file(file)
        # noinspection PyProtectedMember
        res = parser._result
        if res is None:
            raise ValueError("Failed to load Gantt chart.")
        return res


class _GanttParser(LogParser):
    """The log parser for loading Gantt charts."""

    def __init__(self, instance: Instance | None = None):
        """
        Create the gantt parser.

        :param instance: the optional JSSP instance: if `None` is provided,
            we try to load it from the resources
        """
        super().__init__()
        if instance is not None:
            if not isinstance(instance, Instance):
                raise type_error(instance, "instance", Instance)
        #: the internal instance
        self.__instance: Instance | None = instance
        #: the internal section mode: 0=none, 1=setup, 2=y
        self.__sec_mode: int = 0
        #: the gantt string
        self.__gantt_str: str | None = None
        #: the result Gantt chart
        self._result: Gantt | None = None

    def start_section(self, title: str) -> bool:
        """Start a section."""
        super().start_section(title)
        self.__sec_mode = 0
        if title == SECTION_SETUP:
            if self.__instance is None:
                self.__sec_mode = 1
                return True
            return False
        if title == SECTION_RESULT_Y:
            self.__sec_mode = 2
            return True
        return False

    def lines(self, lines: list[str]) -> bool:
        """Parse the lines."""
        if self.__sec_mode == 1:
            if self.__instance is not None:
                raise ValueError(
                    f"instance is already set to {self.__instance}.")
            key: Final[str] = "y.inst.name: "
            for line in lines:
                if line.startswith(key):
                    self.__instance = Instance.from_resource(
                        line[len(key):].strip())
            if self.__instance is None:
                raise ValueError(f"Did not find instance key '{key}' "
                                 f"in section {SECTION_SETUP}!")
        elif self.__sec_mode == 2:
            self.__gantt_str = " ".join(lines).strip()
        else:
            raise ValueError("Should not be in section?")
        return (self.__instance is None) or (self.__gantt_str is None)

    def end_file(self) -> bool:
        """End the file."""
        if self.__gantt_str is None:
            raise ValueError(f"Section {SECTION_RESULT_Y} missing!")
        if self.__instance is None:
            raise ValueError(f"Section {SECTION_SETUP} missing or empty!")
        if self._result is not None:
            raise ValueError("Applied parser to more than one log file?")
        # pylint: disable=C0415,R0401
        from moptipy.examples.jssp.gantt_space import (
            GanttSpace,  # pylint: disable=C0415,R0401
        )
        self._result = GanttSpace(self.__instance).from_str(self.__gantt_str)
        self.__gantt_str = None
        return False



"""Routines for handling strings."""

from math import isnan
from re import MULTILINE, sub
from re import compile as _compile
from typing import Final, Iterable, Pattern

from moptipy.utils.math import __try_int
from moptipy.utils.types import type_error


def float_to_str(x: float) -> str:
    """
    Convert `float` to a string.

    :param x: the floating point value
    :return: the string representation

    >>> float_to_str(1.3)
    '1.3'
    >>> float_to_str(1.0)
    '1'
    """
    if x == 0:
        return "0"
    s = repr(x)
    if isnan(x):
        raise ValueError(f"'{s}' not permitted.")
    if s.endswith(".0"):
        return s[:-2]
    return s


def num_to_str_for_name(x: int | float) -> str:
    """
    Convert a float to a string for use in a component name.

    This function can be inverted by applying :func:`name_str_to_num`.

    :param x: the float
    :returns: the string

    >>> num_to_str_for_name(1.3)
    '1d3'
    >>> num_to_str_for_name(1.0)
    '1'
    >>> num_to_str_for_name(-7)
    'm7'
    >>> num_to_str_for_name(-6.32)
    'm6d32'
    """
    return num_to_str(x).replace(".", DECIMAL_DOT_REPLACEMENT) \
        .replace("-", MINUS_REPLACEMENT)


def name_str_to_num(s: str) -> int | float:
    """
    Convert a string from a name to a number.

    This function is the inverse of :func:`num_to_str_for_name`.

    :param s: the string from the name
    :returns: an integer or float, depending on the number represented by
        `s`

    >>> name_str_to_num(num_to_str_for_name(1.1))
    1.1
    >>> name_str_to_num(num_to_str_for_name(1))
    1
    >>> name_str_to_num(num_to_str_for_name(-5e3))
    -5000
    >>> name_str_to_num(num_to_str_for_name(-6e-3))
    -0.006
    >>> name_str_to_num(num_to_str_for_name(100.0))
    100
    """
    return str_to_intfloat(s.replace(MINUS_REPLACEMENT, "-")
                           .replace(DECIMAL_DOT_REPLACEMENT, "."))


def bool_to_str(value: bool) -> str:
    """
    Convert a Boolean value to a string.

    :param value: the Boolean value
    :return: the string

    >>> print(bool_to_str(True))
    T
    >>> print(bool_to_str(False))
    F
    """
    return "T" if value else "F"


def str_to_bool(value: str) -> bool:
    """
    Convert a string to a boolean value.

    :param value: the string value
    :return: the boolean value

    >>> str_to_bool("T")
    True
    >>> str_to_bool("F")
    False
    >>> try:
    ...     str_to_bool("x")
    ... except ValueError as v:
    ...     print(v)
    Expected 'T' or 'F', but got 'x'.
    """
    if value == "T":
        return True
    if value == "F":
        return False
    raise ValueError(f"Expected 'T' or 'F', but got '{value}'.")


def num_to_str(value: int | float) -> str:
    """
    Transform a numerical type to a string.

    :param value: the value
    :return: the string

    >>> num_to_str(1)
    '1'
    >>> num_to_str(1.5)
    '1.5'
    """
    return str(value) if isinstance(value, int) else float_to_str(value)


def intfloatnone_to_str(val: int | float | None) -> str:
    """
    Convert an integer ot float or `None` to a string.

    :param val: the value
    :return: the string representation
    :rtype: str

    >>> print(repr(intfloatnone_to_str(None)))
    ''
    >>> print(intfloatnone_to_str(12))
    12
    >>> print(intfloatnone_to_str(12.3))
    12.3
    """
    return "" if val is None else num_to_str(val)


def intnone_to_str(val: int | None) -> str:
    """
    Convert an integer or `None` to a string.

    :param val: the value
    :return: the string representation

    >>> print(repr(intnone_to_str(None)))
    ''
    >>> print(intnone_to_str(12))
    12
    """
    return "" if val is None else str(val)


def str_to_intfloat(val: str) -> int | float:
    """
    Convert a string to an int or float.

    :param val: the string value
    :return: the int or float

    >>> print(type(str_to_intfloat("15.0")))
    <class 'int'>
    >>> print(type(str_to_intfloat("15.1")))
    <class 'float'>
    """
    return __try_int(float(val)) if ("e" in val) or \
                                    ("E" in val) or \
                                    ("." in val) or \
                                    ("inf" in val) else int(val)


def str_to_intfloatnone(val: str) -> int | float | None:
    """
    Convert a string to an int or float or None.

    :param val: the string value
    :return: the int or float or None

    >>> print(str_to_intfloatnone(""))
    None
    >>> print(type(str_to_intfloatnone("5.0")))
    <class 'int'>
    >>> print(type(str_to_intfloatnone("5.1")))
    <class 'float'>
    """
    return None if len(val) <= 0 else str_to_intfloat(val)


def str_to_intnone(val: str) -> int | None:
    """
    Convert a string to an int or None.

    :param val: the string value
    :return: the int or None

    >>> print(str_to_intnone(""))
    None
    >>> print(str_to_intnone("5"))
    5
    """
    return None if len(val) <= 0 else int(val)


def replace_all(find: str, replace: str, src: str) -> str:
    """
    Perform a recursive replacement of strings.

    After applying this function, there will not be any occurence of `find`
    left in `src`. All of them will have been replaced by `replace`. If that
    produces new instances of `find`, these will be replaced as well.
    If `replace` contains `find`, this will lead to an endless loop!

    :param find: the string to find
    :param replace: the string with which it will be replaced
    :param src: the string in which we search
    :return: the string `src`, with all occurrences of find replaced by replace

    >>> replace_all("a", "b", "abc")
    'bbc'
    >>> replace_all("aa", "a", "aaaaa")
    'a'
    >>> replace_all("aba", "a", "abaababa")
    'aa'
    """
    new_len = len(src)
    while True:
        src = src.replace(find, replace)
        old_len = new_len
        new_len = len(src)
        if new_len >= old_len:
            return src


def __replace_double(replace: str, src: str) -> str:
    """
    Replace any double-occurrence of a string with a single occurrence.

    :param replace: the string to replace
    :param src: the source string
    :returns: the updated string
    """
    return replace_all(replace + replace, replace, src)


#: the separator of different filename parts
PART_SEPARATOR: Final[str] = "_"
#: the replacement for "." in a file name
DECIMAL_DOT_REPLACEMENT: Final[str] = "d"
#: the replacement for "-" in a file name
MINUS_REPLACEMENT: Final[str] = "m"
#: the replacement for "+" in a file name
PLUS_REPLACEMENT: Final[str] = "p"

#: a pattern used during name sanitization
__PATTERN_SPACE_BEFORE_MINUS: Final[Pattern] = _compile(r"[^\w\s-]")
#: the multiple-whitespace pattern
__PATTERN_MULTIPLE_WHITESPACE: Final[Pattern] = _compile(r"\s+")


def sanitize_name(name: str) -> str:
    """
    Sanitize a name in such a way that it can be used as path component.

    >>> sanitize_name(" hello world ")
    'hello_world'
    >>> sanitize_name(" 56.6-455 ")
    '56d6m455'
    >>> sanitize_name(" _ i _ am _ funny   --6 _ ")
    'i_am_funny_m6'

    :param name: the name that should be sanitized
    :return: the sanitized name
    :raises ValueError: if the name is invalid or empty
    :raises TypeError: if the name is `None` or not a string
    """
    if not isinstance(name, str):
        raise type_error(name, "name", str)
    orig_name = name
    name = name.strip()
    name = __replace_double("-", name).replace("+", PLUS_REPLACEMENT)
    name = __replace_double("+", name).replace("-", MINUS_REPLACEMENT)
    name = __replace_double("_", name)
    name = __replace_double(".", name).replace(".", DECIMAL_DOT_REPLACEMENT)

    name = sub(__PATTERN_SPACE_BEFORE_MINUS, "", name)
    name = sub(__PATTERN_MULTIPLE_WHITESPACE, PART_SEPARATOR, name)
    name = __replace_double("_", name)

    if name.startswith("_"):
        name = name[1:]

    if name.endswith("_"):
        name = name[:len(name) - 1]

    if len(name) <= 0:
        raise ValueError(
            f"Sanitized name must not become empty, but '{orig_name}' does.")

    return name


def sanitize_names(names: Iterable[str]) -> str:
    """
    Sanitize a set of names.

    >>> sanitize_names(["", " sdf ", "", "5-3"])
    'sdf_5m3'
    >>> sanitize_names([" a ", " b", " c", "", "6", ""])
    'a_b_c_6'

    :param names: the list of names.
    :return: the sanitized name
    """
    return PART_SEPARATOR.join([
        sanitize_name(name) for name in names if len(name) > 0])


def regex_sub(search: str | Pattern,
              replace: str,
              inside: str) -> str:
    r"""
    Replace all occurrences of 'search' in 'inside' with 'replace'.

    :param search: the regular expression to search
    :param replace: the regular expression to replace it with
    :param inside: the string in which to search/replace
    :return: the new string after the recursive replacement

    >>> regex_sub('[ \t]+\n', '\n', ' bla \nxyz\tabc\t\n')
    ' bla\nxyz\tabc\n'
    >>> regex_sub('[0-9]A', 'X', '23A7AA')
    '2XXA'
    """
    while True:
        text = sub(search, replace, inside, MULTILINE)
        if text is inside:
            return inside
        inside = text



"""A simple and immutable basic statistics record."""

import statistics
from dataclasses import dataclass
from math import gcd, inf, isfinite, sqrt
from typing import Callable, Final, Iterable, cast

from moptipy.utils.logger import CSV_SEPARATOR, SCOPE_SEPARATOR
from moptipy.utils.math import (
    DBL_INT_LIMIT_P,
    try_int,
    try_int_div,
    try_int_root,
)
from moptipy.utils.strings import (
    num_to_str,
    str_to_intfloat,
    str_to_intfloatnone,
)
from moptipy.utils.types import type_error

#: The limit until which we simplify geometric mean data.
_INT_ROOT_LIMIT: Final[int] = int(sqrt(DBL_INT_LIMIT_P))
_ULP: Final[float] = 1 - (2 ** (-53))

#: The minimum value key.
KEY_MINIMUM: Final[str] = "min"
#: The median value key.
KEY_MEDIAN: Final[str] = "med"
#: The arithmetic mean value key.
KEY_MEAN_ARITH: Final[str] = "mean"
#: The geometric mean value key.
KEY_MEAN_GEOM: Final[str] = "geom"
#: The maximum value key.
KEY_MAXIMUM: Final[str] = "max"
#: The standard deviation value key.
KEY_STDDEV: Final[str] = "sd"

#: The number of CSV columns.
CSV_COLS: Final[int] = 6

#: The empty csv row of statistics
EMPTY_CSV_ROW: Final[str] = CSV_SEPARATOR * (CSV_COLS - 1)

#: the internal getters
_GETTERS: Final[dict[str, Callable[["Statistics"],
                                   int | float | None]]] = {
    KEY_MINIMUM: lambda s: s.minimum,
    KEY_MEDIAN: lambda s: s.median,
    KEY_MEAN_ARITH: lambda s: s.mean_arith,
    KEY_MEAN_GEOM: lambda s: s.mean_geom,
    KEY_MAXIMUM: lambda s: s.maximum,
    KEY_STDDEV: lambda s: s.stddev
}


@dataclass(frozen=True, init=False, order=True)
class Statistics:
    """An immutable record with statistics of one quantity."""

    #: The minimum.
    minimum: int | float
    #: The median.
    median: int | float
    #: The arithmetic mean value.
    mean_arith: int | float
    #: The geometric mean value, if defined.
    mean_geom: int | float | None
    #: The maximum.
    maximum: int | float
    #: The standard deviation.
    stddev: int | float

    def __init__(self, n: int,
                 minimum: int | float,
                 median: int | float,
                 mean_arith: int | float,
                 mean_geom: int | float | None,
                 maximum: int | float,
                 stddev: int | float):
        """
        Initialize the statistics class.

        :param minimum: the minimum
        :param median: the median
        :param mean_arith: the arithmetic mean
        :param mean_geom: the geometric mean, or `None` if it is undefined
        :param maximum: the maximum
        :param stddev: the standard deviation (`0` is also used for undefined)
        """
        if not isinstance(n, int):
            raise type_error(n, "n", int)
        if n <= 0:
            raise ValueError(f"n must be >= 1, but is {n}.")

        # check minimum
        if not isinstance(minimum, (int, float)):
            raise type_error(minimum, "minimum", (int, float))
        if isinstance(minimum, float) and (not isfinite(minimum)):
            raise ValueError(f"minimum must be finite, but is {minimum}.")

        # check median
        if not isinstance(median, (int, float)):
            raise type_error(median, "median", (int, float))
        if isinstance(median, float) and (not isfinite(median)):
            raise ValueError(f"med must be finite, but is {median}.")
        if n == 1:
            if median != minimum:
                raise ValueError(f"median ({median}) must equal "
                                 f"minimum ({minimum}) if n=1.")
        elif median < minimum:
            raise ValueError(
                f"median ({median}) must be >= minimum ({minimum}) if n>1.")

        # check maximum
        if not isinstance(maximum, (int, float)):
            raise type_error(maximum, "maximum", (int, float))
        if isinstance(maximum, float) and (not isfinite(maximum)):
            raise ValueError(f"maximum must be finite, but is {maximum}.")
        if n == 1:
            if maximum != minimum:
                raise ValueError(f"maximum ({maximum}) must equal "
                                 f"minimum ({minimum}) if n=1.")
        elif maximum < median:
            raise ValueError(
                f"maximum ({maximum}) must be >= med ({median}) if n>1.")

        # check arithmetic mean
        if not isinstance(mean_arith, (int, float)):
            raise type_error(mean_arith, "mean_arith", (int, float))
        if isinstance(mean_arith, float) and (not isfinite(mean_arith)):
            raise ValueError(
                f"mean_arith must be finite, but is {mean_arith}.")
        if n == 1:
            if mean_arith != minimum:
                raise ValueError(f"mean_arith ({mean_arith}) must equal "
                                 f"minimum ({minimum}) if n=1.")
        else:
            if mean_arith < minimum:
                raise ValueError(
                    f"mean_arith ({mean_arith}) must be >= "
                    f"minimum ({minimum}) if n>1.")
            if mean_arith > maximum:
                raise ValueError(
                    f"mean_arith ({mean_arith}) must be <= "
                    f"maximum ({maximum}) if n>1.")

        # check geometric mean
        if mean_geom is None:
            if minimum > 0:
                raise ValueError(
                    f"If minimum ({minimum}) > 0, then mean_geom must be "
                    f"defined, but it is {mean_geom}.")
        else:
            if minimum <= 0:
                raise ValueError(
                    f"If minimum ({minimum}) <= 0, then mean_geom is "
                    f"undefined, but it is {mean_geom}.")
            if not isinstance(mean_geom, (int, float)):
                raise type_error(mean_geom, "mean_geom", (int, float))
            if isinstance(mean_geom, float) and (not isfinite(mean_geom)):
                raise ValueError(
                    f"mean_geom must be finite, but is {mean_geom}.")
            if n == 1:
                if mean_geom != minimum:
                    raise ValueError(f"mean_geom ({mean_geom}) must equal "
                                     f"minimum ({minimum}) if n=1.")
            else:
                if mean_geom < minimum:
                    raise ValueError(
                        f"mean_geom ({mean_geom}) must be >= "
                        f"minimum ({minimum}) if n>1.")
                if mean_geom > mean_arith:
                    raise ValueError(
                        f"mean_geom ({mean_geom}) must be <= "
                        f"mean_arith ({mean_arith}) if n>1.")
                if (mean_geom >= mean_arith) and (minimum < maximum):
                    raise ValueError(
                        f"mean_geom ({mean_geom}) must be < "
                        f"mean_arith ({mean_arith}) if n>1 and "
                        f"minimum ({minimum}) < maximum ({maximum}).")

        if not isinstance(stddev, (int, float)):
            raise type_error(stddev, "stddev", (int, float))
        if isinstance(stddev, float) and (not isfinite(stddev)):
            raise ValueError(f"stddev must be finite, but is {stddev}.")

        if n == 1:
            if stddev != 0:
                raise ValueError(
                    f"If n==1, stddev must be 0, but is {stddev}.")
            stddev = 0  # set it to int 0

        if (stddev == 0) ^ (minimum == maximum):
            raise ValueError(
                f"If stddev ({stddev}) is 0, then minimum ({minimum}) must "
                f"equal maximum ({maximum}) and vice versa.")
        if minimum == maximum:
            stddev = 0  # set if to int 0

        # fix types to int where possible without loss of precision

        object.__setattr__(self, "minimum", try_int(minimum))
        object.__setattr__(self, "median", try_int(median))
        object.__setattr__(self, "maximum", try_int(maximum))
        object.__setattr__(self, "mean_arith", try_int(mean_arith))
        object.__setattr__(self, "mean_geom",
                           None if mean_geom is None else try_int(mean_geom))
        object.__setattr__(self, "stddev", try_int(stddev))

    def min_mean(self) -> int | float:
        """
        Obtain the smallest of the three mean values.

        :return: the smallest of `mean_arith`, `mean_geom`, and `median`
        :rtype: Union[int, float]
        """
        if self.mean_geom is None:
            if self.mean_arith < self.median:
                return self.mean_arith
            return self.median

        if self.mean_geom < self.median:
            return self.mean_geom
        return self.median

    def max_mean(self) -> int | float:
        """
        Obtain the largest of the three mean values.

        :return: the largest of `mean_arith`, `mean_geom`, and `median`
        :rtype: Union[int, float]
        """
        if self.mean_arith > self.median:
            return self.mean_arith
        return self.median

    @staticmethod
    def create(source: Iterable[int | float]) -> "Statistics":
        """
        Create a statistics object from an iterable.

        :param source: the source
        :return: a statistics representing the statistics over `source`
        :rtype: Statistics

        >>> from moptipy.evaluation.statistics import Statistics
        >>> s = Statistics.create([3, 1, 2, 5])
        >>> print(s.minimum)
        1
        >>> print(s.maximum)
        5
        >>> print(s.mean_arith)
        2.75
        >>> print(s.median)
        2.5
        >>> print(f"{s.mean_geom:.4f}")
        2.3403
        >>> print(f"{s.min_mean():.4f}")
        2.3403
        >>> print(f"{s.max_mean()}")
        2.75
        """
        if not isinstance(source, Iterable):
            raise type_error(source, "source", Iterable)

        minimum: int | float = inf
        maximum: int | float = -inf
        can_int: bool = True  # are all values integers?
        int_sum: int = 0  # the integer sum (for mean, stddev)
        int_prod: int = 1  # the integer product (for geom_mean)
        int_sum_sqr: int = 0  # the sum of squares (for stddev)
        n: int = 0  # the number of items

        for e in source:  # iterate over all data
            n = n + 1
            e = try_int(e)

            if can_int:  # can we do integers
                if not isinstance(e, int):
                    can_int = False
            if can_int:  # so far, we only encountered ints
                zz = cast(int, e)
                int_sum += zz  # so we can sum exactly
                int_prod *= zz  # and can compute exact products
                int_sum_sqr += zz * zz

            if e < minimum:
                minimum = e  # update minimum
            if e > maximum:
                maximum = e  # update maximum

        if n <= 0:
            raise ValueError("source cannot be empty.")

        if minimum >= maximum:
            return Statistics(n=n,
                              minimum=minimum,
                              median=minimum,
                              mean_arith=minimum,
                              mean_geom=None if minimum <= 0 else minimum,
                              maximum=maximum,
                              stddev=0)

        stddev: int | float = 0
        mean_geom: int | float | None = None
        mean_arith: int | float

        if can_int:  # if we get here, we have exact sums and product
            mean_arith = try_int_div(int_sum, n)

            if n > 1:  # standard deviation only defined for n > 1
                int_sum2: int = (int_sum * int_sum)
                i_gcd = gcd(int_sum2, n)
                int_sum2 = int_sum2 // i_gcd
                i_n = n // i_gcd

                var: int | float  # the container for the variance
                if i_n == 1:
                    var = try_int_div(int_sum_sqr - int_sum2, n - 1)
                else:  # variance is float
                    var = (int_sum_sqr - (int_sum2 / i_n)) / (n - 1)

                stddev = try_int(sqrt(var))

            if minimum > 0:  # geometric mean only defined for all-positive
                if int_prod == 0:
                    mean_geom = 0  # geometric mean is 0 if product is 0
                else:  # if not, geom_mean = prod ** (1/n)
                    mean_geom = try_int_root(int_prod, n, True)
                    if mean_geom is None:
                        mean_geom = statistics.geometric_mean(source)
        else:  # ok, we do not have only integer-like values
            mean_arith = try_int(statistics.mean(source))
            if minimum > 0:
                mean_geom = try_int(statistics.geometric_mean(source))
            if n > 1:
                stddev = try_int(statistics.stdev(source))

        if mean_geom is not None:
            # Deal with errors that may have arisen due to
            # numerical imprecision.
            if mean_geom < minimum:
                if (mean_geom / _ULP) >= (minimum * _ULP):
                    mean_geom = minimum
            if mean_geom > maximum:
                if (maximum / _ULP) >= (mean_geom * _ULP):
                    mean_geom = maximum

        return Statistics(n=n,
                          minimum=minimum,
                          median=statistics.median(source),
                          mean_arith=mean_arith,
                          mean_geom=mean_geom,
                          maximum=maximum,
                          stddev=stddev)

    @staticmethod
    def csv_col_names(prefix: str) -> Iterable[str]:
        """
        Make the column names suitable for a CSV-formatted file.

        :param str prefix: the prefix name of the columns
        :return: the column header strings
        :rtype: Iterable[str]
        """
        return [f"{prefix}{SCOPE_SEPARATOR}{b}"
                for b in [KEY_MINIMUM, KEY_MEDIAN,
                          KEY_MEAN_ARITH, KEY_MEAN_GEOM,
                          KEY_MAXIMUM, KEY_STDDEV]]

    @staticmethod
    def value_to_csv(value: int | float) -> str:
        """
        Expand a single value to a CSV row.

        :param Union[int, float] value: the value
        :return: the CSV row.
        :rtype: str
        """
        s: str = f"{num_to_str(value)}{CSV_SEPARATOR}"
        return f"{s * 5}0"

    def to_csv(self) -> str:
        """
        Generate a string with the data of this record in CSV format.

        :return: the string
        :rtype: str
        """
        q: Final[str] = "" if (self.mean_geom is None) \
            else num_to_str(self.mean_geom)
        return f"{num_to_str(self.minimum)}{CSV_SEPARATOR}" \
               f"{num_to_str(self.median)}{CSV_SEPARATOR}" \
               f"{num_to_str(self.mean_arith)}" \
               f"{CSV_SEPARATOR}" \
               f"{q}" \
               f"{CSV_SEPARATOR}" \
               f"{num_to_str(self.maximum)}{CSV_SEPARATOR}" \
               f"{num_to_str(self.stddev)}"

    @staticmethod
    def from_csv(n: int, row: str | Iterable[str]) -> "Statistics":
        """
        Convert a CSV string (or separate CSV cells) to a Statistics object.

        :param int n: the number of observations
        :param Union[str, Iterable[str]] row: either the single string or
            the iterable of separate strings
        :return: the :class:`Statistics` instance
        :rtype: Statistics
        """
        cells = row.split(CSV_SEPARATOR) \
            if isinstance(row, str) else row

        mini, med, mean, geo, maxi, sd = cells
        return Statistics(n,
                          str_to_intfloat(mini),
                          str_to_intfloat(med),
                          str_to_intfloat(mean),
                          str_to_intfloatnone(geo),
                          str_to_intfloat(maxi),
                          str_to_intfloat(sd))

    @staticmethod
    def getter(dimension: str) -> Callable[["Statistics"],
                                           int | float | None]:
        """
        Produce a function that obtains the given dimension from Statistics.

        :param dimension: the dimension
        :returns: a callable that returns the value corresponding to the
            dimension
        """
        if not isinstance(dimension, str):
            raise type_error(dimension, "dimension", str)
        if dimension in _GETTERS:
            return _GETTERS[dimension]
        raise ValueError(f"unknown dimension '{dimension}', "
                         f"should be one of {sorted(_GETTERS.keys())}.")



"""
Fitness assignment processes.

Fitness assignment processes are modules of the fully-configurable
Evolutionary Algorithm :class:`~moptipy.algorithms.so.general_ea.GeneralEA`.
The transform objective values to scalar fitness values that are then used
by :class:`~moptipy.algorithms.modules.selection.Selection` algorithms.
"""



"""Statistics aggregated over multiple instances of EndResult."""
import os.path
import sys
from dataclasses import dataclass
from math import ceil, inf
from typing import Any, Callable, Final, Iterable, Union

import moptipy.api.logging as log
from moptipy.evaluation._utils import _check_max_time_millis
from moptipy.evaluation.base import (
    F_NAME_RAW,
    F_NAME_SCALED,
    KEY_N,
    MultiRunData,
)
from moptipy.evaluation.end_results import EndResult
from moptipy.evaluation.statistics import (
    CSV_COLS,
    EMPTY_CSV_ROW,
    KEY_STDDEV,
    Statistics,
)
from moptipy.utils.console import logger
from moptipy.utils.help import help_screen
from moptipy.utils.logger import CSV_SEPARATOR, SCOPE_SEPARATOR
from moptipy.utils.math import try_int, try_int_div
from moptipy.utils.path import Path
from moptipy.utils.strings import num_to_str, sanitize_name, str_to_intfloat
from moptipy.utils.types import type_error, type_name_of

#: The key for the best F.
KEY_BEST_F_SCALED: Final[str] = log.KEY_BEST_F + "scaled"
#: The key for the number of successful runs.
KEY_N_SUCCESS: Final[str] = "successN"
#: The key for the success FEs.
KEY_SUCCESS_FES: Final[str] = "successFEs"
#: The key for the success time millis.
KEY_SUCCESS_TIME_MILLIS: Final[str] = "successTimeMillis"
#: The key for the ERT in FEs.
KEY_ERT_FES: Final[str] = "ertFEs"
#: The key for the ERT in milliseconds.
KEY_ERT_TIME_MILLIS: Final[str] = "ertTimeMillis"

#: the internal getters that can work directly
_GETTERS_0: Final[dict[str, Callable[["EndStatistics"],
                                     int | float | None]]] = {
    KEY_N_SUCCESS: lambda s: s.n_success,
    KEY_ERT_FES: lambda s: s.ert_fes,
    KEY_ERT_TIME_MILLIS: lambda s: s.ert_time_millis,
    log.KEY_GOAL_F: lambda s: s.goal_f if isinstance(s.goal_f,
                                                     (int, float)) else None,
    log.KEY_MAX_TIME_MILLIS: lambda s: s.max_time_millis
    if isinstance(s.max_time_millis, (int, float)) else None,
    log.KEY_MAX_FES: lambda s: s.max_fes
    if isinstance(s.max_fes, (int, float)) else None,
}

#: the internal getters that access end statistics
_GETTERS_1: Final[dict[
    str, Callable[["EndStatistics"], int | float | Statistics | None]]] = {
    log.KEY_LAST_IMPROVEMENT_FE: lambda s: s.last_improvement_fe,
    log.KEY_LAST_IMPROVEMENT_TIME_MILLIS:
        lambda s: s.last_improvement_time_millis,
    log.KEY_TOTAL_FES: lambda s: s.total_fes,
    log.KEY_TOTAL_TIME_MILLIS: lambda s: s.total_time_millis,
    F_NAME_RAW: lambda s: s.best_f,
    F_NAME_SCALED: lambda s: s.best_f_scaled,
    log.KEY_MAX_TIME_MILLIS: lambda s: s.max_time_millis,
    log.KEY_MAX_FES: lambda s: s.max_fes,
    log.KEY_GOAL_F: lambda s: s.goal_f,
}
_GETTERS_1[log.KEY_BEST_F] = _GETTERS_1[F_NAME_RAW]
_GETTERS_1[KEY_BEST_F_SCALED] = _GETTERS_1[F_NAME_SCALED]


@dataclass(frozen=True, init=False, order=True)
class EndStatistics(MultiRunData):
    """
    Statistics over end results of one or multiple algorithm*instance setups.

    If one algorithm*instance is used, then `algorithm` and `instance` are
    defined. Otherwise, only the parameter which is the same over all recorded
    runs is defined.
    """

    #: The statistics about the best encountered result.
    best_f: Statistics
    #: The statistics about the last improvement FE.
    last_improvement_fe: Statistics
    #: The statistics about the last improvement time.
    last_improvement_time_millis: Statistics
    #: The statistics about the total number of FEs.
    total_fes: Statistics
    #: The statistics about the total time.
    total_time_millis: Statistics
    #: The goal objective value.
    goal_f: Statistics | int | float | None
    #: best_f / goal_f if goal_f is consistently defined and always positive.
    best_f_scaled: Statistics | None
    #: The number of successful runs, if goal_f != None, else None.
    n_success: int | None
    #: The FEs to success, if n_success > 0, None otherwise.
    success_fes: Statistics | None
    #: The time to success, if n_success > 0, None otherwise.
    success_time_millis: Statistics | None
    #: The ERT if FEs, while is inf if n_success=0, None if goal_f is None,
    #: and finite otherwise.
    ert_fes: int | float | None
    #: The ERT if milliseconds, while is inf if n_success=0, None if goal_f
    #: is None, and finite otherwise.
    ert_time_millis: int | float | None
    #: The budget in FEs, if every run had one; None otherwise.
    max_fes: Statistics | int | None
    #: The budget in milliseconds, if every run had one; None otherwise.
    max_time_millis: Statistics | int | None

    def __init__(self,
                 algorithm: str | None,
                 instance: str | None,
                 n: int,
                 best_f: Statistics,
                 last_improvement_fe: Statistics,
                 last_improvement_time_millis: Statistics,
                 total_fes: Statistics,
                 total_time_millis: Statistics,
                 goal_f: float | int | Statistics | None,
                 best_f_scaled: Statistics | None,
                 n_success: int | None,
                 success_fes: Statistics | None,
                 success_time_millis: Statistics | None,
                 ert_fes: int | float | None,
                 ert_time_millis: int | float | None,
                 max_fes: Statistics | int | None,
                 max_time_millis: Statistics | int | None):
        """
        Create the end statistics of an experiment-setup combination.

        :param algorithm: the algorithm name, if all runs are with the same
            algorithm
        :param instance: the instance name, if all runs are on the same
            instance
        :param n: the total number of runs
        :param best_f: statistics about the best achieved result
        :param last_improvement_fe: statistics about the last improvement FE
        :param last_improvement_time_millis: statistics about the last
            improvement time
        :param total_fes: statistics about the total FEs
        :param total_time_millis: statistics about the total runtime in
            milliseconds
        :param goal_f: if the goal objective value is not defined sometimes,
            this will be `None`. If it is always defined and always the same,
            then this will be that value. If different goal values exist, then
            this is the `Statistics` record about them
        :param best_f_scaled: if `goal_f` is not `None` and greater than zero,
            then here we provide statistics about `best_f` divided by the
            corresponding `goal_f`
        :param n_success: the number of successful runs is only defined if
            `goal_f` is not `None` and counts the number of runs that reach or
            surpass their corresponding `goal_f`
        :param success_fes: if `goal_f` is not `None`,
            then this holds statistics about the last improvement FE of only
            the successful runs
        :param success_time_millis: if `goal_f` is not `None`, then this holds
            statistics about the last improvement times of only the successful
            runs
        :param ert_fes: if `goal_f` is always defined, then this is the
            empirically estimated running time to solve the problem in FEs if
            `n_success>0` and `inf` otherwise
        :param ert_time_millis: if `goal_f` is always defined, then this is
            the empirically estimated running time to solve the problem in
            milliseconds if `n_success>0` and `inf` otherwise
        :param max_fes: the budget in FEs, if any
        :param max_time_millis: the budget in term of milliseconds
        """
        super().__init__(algorithm, instance, n)

        if not isinstance(best_f, Statistics):
            raise type_error(best_f, "best_f", Statistics)
        object.__setattr__(self, "best_f", best_f)

        if not isinstance(last_improvement_fe, Statistics):
            raise type_error(last_improvement_fe, "last_improvement_fe",
                             Statistics)
        if last_improvement_fe.minimum <= 0:
            raise ValueError("No last_improvement_fe can be <= 0, but "
                             f"encountered {last_improvement_fe.minimum}.")
        if not isinstance(last_improvement_fe.minimum, int):
            raise type_error(last_improvement_fe.minimum,
                             "minimum last_improvement_fe", int)
        if not isinstance(last_improvement_fe.maximum, int):
            raise type_error(last_improvement_fe.maximum,
                             "maximum last_improvement_fe", int)
        object.__setattr__(self, "last_improvement_fe", last_improvement_fe)

        if not isinstance(last_improvement_time_millis, Statistics):
            raise type_error(last_improvement_time_millis,
                             "last_improvement_time_millis", Statistics)
        object.__setattr__(self, "last_improvement_time_millis",
                           last_improvement_time_millis)
        if last_improvement_time_millis.minimum < 0:
            raise ValueError(
                "No last_improvement_time_millis can be < 0, but encountered "
                f"{last_improvement_time_millis.minimum}.")
        if not isinstance(last_improvement_time_millis.minimum, int):
            raise type_error(last_improvement_time_millis.minimum,
                             "minimum last_improvement_time_millis", int)
        if not isinstance(last_improvement_time_millis.maximum, int):
            raise type_error(last_improvement_time_millis.maximum,
                             "maximum last_improvement_time_millis", int)

        if not isinstance(total_fes, Statistics):
            raise type_error(total_fes, "total_fes", Statistics)
        total_fes_min: Final[int | float] = total_fes.minimum
        total_fes_max: Final[int | float] = total_fes.maximum
        if total_fes_min <= 0:
            raise ValueError("No total_fes can be <= 0, but "
                             f"encountered {total_fes_min}.")
        if not isinstance(total_fes_min, int):
            raise type_error(total_fes_min, "minimum total_fes", int)
        if not isinstance(total_fes_max, int):
            raise type_error(total_fes_max, "maximum total_fes", int)
        if total_fes_min < last_improvement_fe.minimum:
            raise ValueError(
                f"Minimum total_fes ({total_fes_min}) cannot be"
                "less than minimum last_improvement_fe "
                f"({last_improvement_fe.minimum}).")
        if total_fes_max < last_improvement_fe.maximum:
            raise ValueError(
                f"Maximum total_fes ({total_fes_max}) cannot be"
                "less than maximum last_improvement_fe "
                f"({last_improvement_fe.maximum}).")
        object.__setattr__(self, "total_fes", total_fes)

        if not isinstance(total_time_millis, Statistics):
            raise type_error(total_time_millis, "total_time_millis",
                             Statistics)
        total_time_min: Final[int | float] = total_time_millis.minimum
        total_time_max: Final[int | float] = total_time_millis.maximum

        if total_time_min < 0:
            raise ValueError("No total_time_millis can be < 0, but "
                             f"encountered {total_time_min}.")
        if not isinstance(total_time_min, int):
            raise type_error(total_time_min, "minimum total_time_millis", int)
        if not isinstance(total_time_max, int):
            raise type_error(total_time_max, "maximum total_time_millis", int)
        if total_time_min < last_improvement_time_millis.minimum:
            raise ValueError(
                f"Minimum total_time_millis ({total_time_min}) "
                "cannot be less than minimum last_improvement_time_millis "
                f"({last_improvement_time_millis.minimum}).")
        if total_time_max < last_improvement_time_millis.maximum:
            raise ValueError(
                f"Maximum total_time_millis ({total_time_max}) "
                "cannot be less than maximum last_improvement_time_millis "
                f"({last_improvement_time_millis.maximum}).")
        object.__setattr__(self, "total_time_millis", total_time_millis)

        if goal_f is None:
            if best_f_scaled is not None:
                raise ValueError(
                    "If goal_f is None, best_f_scaled must also be None, "
                    f"but is {type(best_f_scaled)}.")
            if n_success is not None:
                raise ValueError(
                    "If goal_f is None, n_success must also be None, "
                    f"but is {type(n_success)}.")
            if success_fes is not None:
                raise ValueError(
                    "If success_fes is None, best_f_scaled must also be None, "
                    f"but is {type(success_fes)}.")
            if success_time_millis is not None:
                raise ValueError(
                    "If success_time_millis is None, best_f_scaled "
                    "must also be None, "
                    f"but is {type(success_time_millis)}.")
            if ert_fes is not None:
                raise ValueError(
                    "If goal_f is None, ert_fes must also be None, "
                    f"but is {type(ert_fes)}.")
            if ert_time_millis is not None:
                raise ValueError(
                    "If goal_f is None, ert_time_millis must also be None, "
                    f"but is {type(ert_time_millis)}.")
        else:  # goal_f is not None
            if isinstance(goal_f, float):
                if goal_f <= (-inf):
                    goal_f = None
                else:
                    goal_f = try_int(goal_f)
            elif not isinstance(goal_f, (int, Statistics)):
                raise type_error(goal_f, "goal_f", (int, Statistics))

            if best_f_scaled is not None:
                goal_f_min: Final[int | float] = \
                    goal_f.minimum if isinstance(goal_f, Statistics) \
                    else goal_f
                if goal_f_min <= 0:
                    raise ValueError(
                        f"best_f_scaled must be None if minimum goal_f "
                        f"({goal_f_min}) of goal_f {goal_f} is not positive,"
                        f" but is {best_f_scaled}.")
                if not isinstance(best_f_scaled, Statistics):
                    raise type_error(best_f_scaled, "best_f_scaled",
                                     Statistics)
                if best_f_scaled.minimum < 0:
                    raise ValueError(
                        "best_f_scaled cannot be negative, but encountered "
                        f"{best_f_scaled.minimum}.")

            if not isinstance(n_success, int):
                raise type_error(n_success, "n_success", int)
            if n_success < 0:
                raise ValueError(
                    f"n_success must be positive, but is {n_success}.")

            if not isinstance(ert_fes, (int, float)):
                raise type_error(ert_fes, "ert_fes", (int, float))
            if not isinstance(ert_time_millis, (int, float)):
                raise type_error(ert_time_millis, "ert_time_millis",
                                 (int, float))

            if n_success > 0:
                if not isinstance(success_fes, Statistics):
                    raise type_error(success_fes,
                                     "if n_success>0, then success_fes",
                                     Statistics)
                if not isinstance(success_fes.minimum, int):
                    raise type_error(success_fes.minimum,
                                     "success_fes.minimum", int)
                if not isinstance(success_fes.maximum, int):
                    raise type_error(success_fes.maximum,
                                     "success_fes.maximum", int)
                if success_fes.minimum < last_improvement_fe.minimum:
                    raise ValueError(
                        "success_fes.minimum must be >= "
                        f"{last_improvement_fe.minimum}, but is "
                        f"{success_fes.minimum}.")
                if success_fes.maximum > last_improvement_fe.maximum:
                    raise ValueError(
                        "success_fes.maximum must be <= "
                        f"{last_improvement_fe.maximum}, but is "
                        f"{success_fes.maximum}.")
                if not isinstance(success_time_millis, Statistics):
                    raise type_error(
                        success_time_millis,
                        "if n_success>0, then success_time_millis",
                        Statistics)
                if not isinstance(success_time_millis.minimum, int):
                    raise type_error(success_time_millis.minimum,
                                     "success_time_millis.minimum", int)
                if not isinstance(success_time_millis.maximum, int):
                    raise type_error(success_time_millis.maximum,
                                     "success_time_millis.maximum", int)
                if success_time_millis.minimum < \
                        last_improvement_time_millis.minimum:
                    raise ValueError(
                        "success_time_millis.minimum must be >= "
                        f"{last_improvement_time_millis.minimum}, but is "
                        f"{success_time_millis.minimum}.")
                if success_time_millis.maximum > \
                        last_improvement_time_millis.maximum:
                    raise ValueError(
                        "success_time_millis.maximum must be <= "
                        f"{last_improvement_time_millis.maximum}, but is "
                        f"{success_time_millis.maximum}.")

                ert_fes = try_int(ert_fes)
                if ert_fes < success_fes.minimum:
                    raise ValueError(
                        "ert_fes must be >= "
                        f"{success_fes.minimum}, but is {ert_fes}.")
                ert_fe_max = ceil(total_fes.mean_arith * n)
                if ert_fes > ert_fe_max:
                    raise ValueError(
                        "ert_fes must be <= "
                        f"{ert_fe_max}, but is {ert_fes}.")

                ert_time_millis = try_int(ert_time_millis)
                if ert_time_millis < success_time_millis.minimum:
                    raise ValueError(
                        "ert_time_millis must be >= "
                        f"{success_time_millis.minimum}, but "
                        f"is {ert_time_millis}.")
                ert_time_max = ceil(total_time_millis.mean_arith * n)
                if ert_time_millis > ert_time_max:
                    raise ValueError(
                        "ert_time_millis must be <= "
                        f"{ert_time_max}, but is {ert_time_millis}.")
            else:
                if success_fes is not None:
                    raise ValueError(
                        "If n_success<=0, then success_fes must be None, "
                        f"but it's a {type_name_of(success_fes)}.")
                if success_time_millis is not None:
                    raise ValueError(
                        "If n_success<=0, then success_time_millis must be "
                        f"None, but it is a "
                        f"{type_name_of(success_time_millis)}.")
                if ert_fes < inf:
                    raise ValueError(
                        "If n_success<=0, then ert_fes must "
                        f"be inf, but it's {ert_fes}.")
                if ert_time_millis < inf:
                    raise ValueError(
                        "If n_success<=0, then ert_time_millis must "
                        f"be inf, but it's {ert_time_millis}.")

        object.__setattr__(self, "goal_f", goal_f)
        object.__setattr__(self, "best_f_scaled", best_f_scaled)
        object.__setattr__(self, "n_success", n_success)
        object.__setattr__(self, "success_fes", success_fes)
        object.__setattr__(self, "success_time_millis", success_time_millis)
        object.__setattr__(self, "ert_fes", ert_fes)
        object.__setattr__(self, "ert_time_millis", ert_time_millis)

        if isinstance(max_fes, int):
            if max_fes < total_fes_max:
                raise ValueError(
                    f"max_fes must be >= {total_fes_max}, but is {max_fes}.")
        elif isinstance(max_fes, Statistics):
            if max_fes.minimum < total_fes_min:
                raise ValueError(
                    f"max_fes.minimum must be >= {total_fes_min},"
                    f" but is {max_fes.minimum}.")
            if max_fes.maximum < total_fes_max:
                raise ValueError(
                    f"max_fes.maximum must be >= {total_fes_max},"
                    f" but is {max_fes.maximum}.")
        elif max_fes is not None:
            raise type_error(max_fes, "max_fes", (int, Statistics, None))
        object.__setattr__(self, "max_fes", max_fes)

        if isinstance(max_time_millis, int):
            _check_max_time_millis(max_time_millis,
                                   total_fes_min,
                                   total_time_max)
        elif isinstance(max_time_millis, Statistics):
            _check_max_time_millis(max_time_millis.minimum,
                                   total_fes_min,
                                   total_time_min)
            _check_max_time_millis(max_time_millis.maximum,
                                   total_fes_min,
                                   total_time_max)
        elif max_time_millis is not None:
            raise type_error(max_time_millis, "max_time_millis",
                             (int, Statistics, None))
        object.__setattr__(self, "max_time_millis", max_time_millis)

    @staticmethod
    def create(source: Iterable[EndResult]) -> "EndStatistics":
        """
        Create an `EndStatistics` Record from an Iterable of `EndResult`.

        :param source: the source
        :return: the statistics
        :rtype: EndStatistics
        """
        if not isinstance(source, Iterable):
            raise type_error(source, "source", Iterable)

        n: int = 0
        best_f: list[int | float] = []
        last_improvement_fe: list[int] = []
        last_improvement_time_millis: list[int] = []
        total_fes: list[int] = []
        total_time_millis: list[int] = []
        max_fes: list[int] | None = []
        max_fes_same: bool = True
        max_time_millis: list[int] | None = []
        max_time_same: bool = True
        goal_f: list[int | float] | None = []
        goal_f_same: bool = True
        best_f_scaled: list[float] | None = []
        n_success: int | None = 0
        success_fes: list[int] | None = []
        success_times: list[int] | None = []

        fes: int = 0
        time: int = 0
        algorithm: str | None = None
        instance: str | None = None

        for er in source:
            if not isinstance(er, EndResult):
                raise type_error(er, "end result", EndResult)
            if n == 0:
                algorithm = er.algorithm
                instance = er.instance
            else:
                if algorithm != er.algorithm:
                    algorithm = None
                if instance != er.instance:
                    instance = None
            n += 1
            best_f.append(er.best_f)
            last_improvement_fe.append(er.last_improvement_fe)
            last_improvement_time_millis.append(
                er.last_improvement_time_millis)
            total_fes.append(er.total_fes)
            total_time_millis.append(er.total_time_millis)
            if er.max_fes is None:
                max_fes = None
            elif max_fes is not None:
                if n > 1:
                    max_fes_same = max_fes_same \
                        and (max_fes[-1] == er.max_fes)
                max_fes.append(er.max_fes)
            if er.max_time_millis is None:
                max_time_millis = None
            elif max_time_millis is not None:
                if n > 1:
                    max_time_same = \
                        max_time_same \
                        and (max_time_millis[-1] == er.max_time_millis)
                max_time_millis.append(er.max_time_millis)

            if er.goal_f is None:
                goal_f = None
                best_f_scaled = None
                n_success = None
                success_fes = None
                success_times = None
            else:
                if goal_f is not None:
                    if n > 1:
                        goal_f_same = goal_f_same and (goal_f[-1] == er.goal_f)
                    goal_f.append(er.goal_f)

                    if er.goal_f <= 0:
                        best_f_scaled = None
                    else:
                        if best_f_scaled is not None:
                            best_f_scaled.append(er.best_f / er.goal_f)

                    if er.best_f <= er.goal_f:
                        n_success += 1
                        success_fes.append(er.last_improvement_fe)
                        success_times.append(er.last_improvement_time_millis)
                        fes += er.last_improvement_fe
                        time += er.last_improvement_time_millis
                    else:
                        fes += er.total_fes
                        time += er.total_time_millis
        if n <= 0:
            raise ValueError("There must be at least one end result record.")

        return EndStatistics(
            algorithm,
            instance,
            n,
            Statistics.create(best_f),
            Statistics.create(last_improvement_fe),
            Statistics.create(last_improvement_time_millis),
            Statistics.create(total_fes),
            Statistics.create(total_time_millis),
            None if (goal_f is None)
            else (goal_f[0] if goal_f_same else Statistics.create(goal_f)),
            None if (best_f_scaled is None)
            else Statistics.create(best_f_scaled),
            n_success,
            None if (n_success is None) or (n_success <= 0)
            else Statistics.create(success_fes),
            None if (n_success is None) or (n_success <= 0)
            else Statistics.create(success_times),
            None if (n_success is None)
            else (inf if (n_success <= 0) else try_int_div(fes, n_success)),
            None if (n_success is None) else
            (inf if (n_success <= 0) else try_int_div(time, n_success)),
            None if max_fes is None else
            (max_fes[0] if max_fes_same else Statistics.create(max_fes)),
            None if max_time_millis is None
            else (max_time_millis[0] if max_time_same
                  else Statistics.create(max_time_millis)))

    @staticmethod
    def from_end_results(source: Iterable[EndResult],
                         consumer: Callable[["EndStatistics"], Any],
                         join_all_algorithms: bool = False,
                         join_all_instances: bool = False) -> None:
        """
        Aggregate statistics over a stream of end results.

        :param source: the stream of end results
        :param consumer: the destination to which the new records will be
            sent, can be the `append` method of a :class:`list`
        :param join_all_algorithms: should the statistics be aggregated
            over all algorithms
        :param join_all_instances: should the statistics be aggregated
            over all algorithms
        """
        if not isinstance(source, Iterable):
            raise type_error(source, "source", Iterable)
        if not callable(consumer):
            raise type_error(consumer, "consumer", call=True)
        if not isinstance(join_all_algorithms, bool):
            raise type_error(join_all_algorithms,
                             "join_all_algorithms", bool)
        if not isinstance(join_all_instances, bool):
            raise type_error(join_all_instances, "join_all_instances", bool)

        if join_all_algorithms and join_all_instances:
            consumer(EndStatistics.create(source))
            return

        sorter: dict[str, list[EndResult]] = {}
        for er in source:
            if not isinstance(er, EndResult):
                raise type_error(source, "end results from source",
                                 EndResult)
            key = er.instance if join_all_algorithms else \
                er.algorithm if join_all_instances else \
                f"{er.algorithm}/{er.instance}"
            if key in sorter:
                lst = sorter[key]
            else:
                lst = []
                sorter[key] = lst
            lst.append(er)

        if len(sorter) <= 0:
            raise ValueError("source must not be empty")

        if len(sorter) > 1:
            keyz = list(sorter.keys())
            keyz.sort()
            for key in keyz:
                consumer(EndStatistics.create(sorter[key]))
        else:
            consumer(EndStatistics.create(
                next(iter(sorter.values()))))

    @staticmethod
    def to_csv(  # noqa
            data: Union["EndStatistics", Iterable["EndStatistics"]],  # noqa
            file: str) -> Path:  # noqa
        """
        Store a set of :class:`EndStatistics` in a CSV file.

        :param data: the data to store
        :param file: the file to generate
        :return: the path to the generated CSV file
        """
        path: Final[Path] = Path.path(file)
        logger(f"Writing end result statistics to CSV file '{path}'.")
        Path.path(os.path.dirname(path)).ensure_dir_exists()

        has_algorithm: bool = False  # 1
        has_instance: bool = False  # 2
        has_goal_f: int = 0  # 4
        has_best_f_scaled: bool = False  # 8
        has_n_success: bool = False  # 16
        has_success_fes: bool = False  # 32
        has_success_time_millis: bool = False  # 64
        has_ert_fes: bool = False  # 128
        has_ert_time_millis: bool = False  # 256
        has_max_fes: int = 0  # 512
        has_max_time_millis: int = 0  # 1024
        checker: int = 2047

        if isinstance(data, EndStatistics):
            data = [data]

        for es in data:
            if es.algorithm is not None:
                has_algorithm = True
                checker &= ~1
            if es.instance is not None:
                has_instance = True
                checker &= ~2
            if es.goal_f is not None:
                if isinstance(es.goal_f, Statistics):
                    has_goal_f = 2
                    checker &= ~4
                elif has_goal_f == 0:
                    has_goal_f = 1
            if es.best_f_scaled is not None:
                has_best_f_scaled = True
                checker &= ~8
            if es.n_success is not None:
                has_n_success = True
                checker &= ~8
            if es.success_fes is not None:
                has_success_fes = True
                checker &= ~32
            if es.success_time_millis is not None:
                has_success_time_millis = True
                checker &= ~64
            if es.ert_fes is not None:
                has_ert_fes = True
                checker &= ~128
            if es.ert_time_millis is not None:
                has_ert_time_millis = True
                checker &= ~256
            if es.max_fes is not None:
                if isinstance(es.max_fes, Statistics):
                    has_max_fes = 2
                    checker &= ~512
                elif has_max_fes == 0:
                    has_max_fes = 1
            if es.max_time_millis is not None:
                if isinstance(es.max_time_millis, Statistics):
                    has_max_time_millis = 2
                    checker &= ~512
                elif has_max_time_millis == 0:
                    has_max_time_millis = 1
            if checker == 0:
                break

        with path.open_for_write() as out:
            wrt: Final[Callable] = out.write
            sep: Final[str] = CSV_SEPARATOR
            if has_algorithm:
                wrt(log.KEY_ALGORITHM)
                wrt(sep)
            if has_instance:
                wrt(log.KEY_INSTANCE)
                wrt(sep)

            def h(p) -> None:
                wrt(sep.join(Statistics.csv_col_names(p)))

            wrt(KEY_N)
            wrt(sep)
            h(log.KEY_BEST_F)
            wrt(sep)
            h(log.KEY_LAST_IMPROVEMENT_FE)
            wrt(sep)
            h(log.KEY_LAST_IMPROVEMENT_TIME_MILLIS)
            wrt(sep)
            h(log.KEY_TOTAL_FES)
            wrt(sep)
            h(log.KEY_TOTAL_TIME_MILLIS)
            if has_goal_f == 1:
                wrt(sep)
                wrt(log.KEY_GOAL_F)
            elif has_goal_f == 2:
                wrt(sep)
                h(log.KEY_GOAL_F)
            if has_best_f_scaled:
                wrt(sep)
                h(KEY_BEST_F_SCALED)
            if has_n_success:
                wrt(sep)
                wrt(KEY_N_SUCCESS)
            if has_success_fes:
                wrt(sep)
                h(KEY_SUCCESS_FES)
            if has_success_time_millis:
                wrt(sep)
                h(KEY_SUCCESS_TIME_MILLIS)
            if has_ert_fes:
                wrt(sep)
                wrt(KEY_ERT_FES)
            if has_ert_time_millis:
                wrt(sep)
                wrt(KEY_ERT_TIME_MILLIS)
            if has_max_fes == 1:
                wrt(sep)
                wrt(log.KEY_MAX_FES)
            elif has_max_fes == 2:
                wrt(sep)
                h(log.KEY_MAX_FES)
            if has_max_time_millis == 1:
                wrt(sep)
                wrt(log.KEY_MAX_TIME_MILLIS)
            elif has_max_time_millis == 2:
                wrt(sep)
                h(log.KEY_MAX_TIME_MILLIS)
            out.write("\n")

            csv: Final[Callable] = Statistics.value_to_csv
            num: Final[Callable] = num_to_str

            for er in data:
                if has_algorithm:
                    if er.algorithm is not None:
                        wrt(er.algorithm)
                    wrt(sep)
                if has_instance:
                    if er.instance is not None:
                        wrt(er.instance)
                    wrt(sep)
                wrt(str(er.n))
                wrt(sep)
                wrt(er.best_f.to_csv())
                wrt(sep)
                wrt(er.last_improvement_fe.to_csv())
                wrt(sep)
                wrt(er.last_improvement_time_millis.to_csv())
                wrt(sep)
                wrt(er.total_fes.to_csv())
                wrt(sep)
                wrt(er.total_time_millis.to_csv())
                if has_goal_f == 1:
                    wrt(sep)
                    if er.goal_f is not None:
                        wrt(num(er.goal_f))
                elif has_goal_f == 2:
                    wrt(sep)
                    if isinstance(er.goal_f, Statistics):
                        wrt(er.goal_f.to_csv())
                    elif isinstance(er.goal_f, (int, float)):
                        wrt(csv(er.goal_f))
                    else:
                        wrt(EMPTY_CSV_ROW)
                if has_best_f_scaled:
                    wrt(sep)
                    if er.best_f_scaled is None:
                        wrt(EMPTY_CSV_ROW)
                    else:
                        wrt(er.best_f_scaled.to_csv())
                if has_n_success:
                    wrt(sep)
                    if er.n_success is not None:
                        wrt(str(er.n_success))
                if has_success_fes:
                    wrt(sep)
                    if er.success_fes is None:
                        wrt(EMPTY_CSV_ROW)
                    else:
                        wrt(er.success_fes.to_csv())
                if has_success_time_millis:
                    wrt(sep)
                    if er.success_time_millis is None:
                        wrt(EMPTY_CSV_ROW)
                    else:
                        wrt(er.success_time_millis.to_csv())
                if has_ert_fes:
                    wrt(sep)
                    if er.ert_fes is not None:
                        wrt(num(er.ert_fes))
                if has_ert_time_millis:
                    wrt(sep)
                    if er.ert_time_millis is not None:
                        wrt(num(er.ert_time_millis))
                if has_max_fes == 1:
                    wrt(sep)
                    if er.max_fes is not None:
                        wrt(str(er.max_fes))
                elif has_max_fes == 2:
                    wrt(sep)
                    if isinstance(er.max_fes, Statistics):
                        wrt(er.max_fes.to_csv())
                    elif isinstance(er.max_fes, (int, float)):
                        wrt(csv(er.max_fes))
                    else:
                        wrt(EMPTY_CSV_ROW)
                if has_max_time_millis == 1:
                    wrt(sep)
                    if er.max_time_millis is not None:
                        wrt(str(er.max_time_millis))
                elif has_max_time_millis == 2:
                    wrt(sep)
                    if isinstance(er.max_time_millis, Statistics):
                        wrt(er.max_time_millis.to_csv())
                    elif isinstance(er.max_time_millis, (int, float)):
                        wrt(csv(er.max_time_millis))
                    else:
                        wrt(EMPTY_CSV_ROW)
                out.write("\n")

        logger(f"Done writing end result statistics to CSV file '{path}'.")
        path.enforce_file()
        return path

    @staticmethod
    def from_csv(file: str,
                 consumer: Callable[["EndStatistics"], Any]) -> None:
        """
        Parse a CSV file and collect all encountered :class:`EndStatistics`.

        :param file: the file to parse
        :param consumer: the consumer to receive all the parsed instances of
            :class:`~moptipy.evaluation.end_statistics.EndStatistics`, can be
            the `append` method of a :class:`list`
        """
        path: Final[Path] = Path.file(file)
        logger(f"Begin reading end result statistics from CSV file '{path}'.")

        sep: Final[str] = CSV_SEPARATOR
        with path.open_for_read() as rd:
            headerrow: Final[list[str]] = rd.readlines(1)
            if (headerrow is None) or (len(headerrow) <= 0):
                raise ValueError(f"No line in file '{file}'.")
            headerstr: Final[str] = headerrow[0].strip()
            header: Final[list[str]] = [ss.strip()
                                        for ss in headerstr.split(sep)]
            if len(header) <= 3:
                raise ValueError(
                    f"Invalid header '{headerstr}' in file '{file}'.")

            idx = 0
            has_algorithm: bool
            if header[0] == log.KEY_ALGORITHM:
                has_algorithm = True
                idx = 1
            else:
                has_algorithm = False

            has_instance: bool
            if header[idx] == log.KEY_INSTANCE:
                has_instance = True
                idx += 1
            else:
                has_instance = False

            csv: Final[Callable] = Statistics.csv_col_names

            if header[idx] != KEY_N:
                raise ValueError(
                    f"Expected to find {KEY_N} at index {idx} "
                    f"in header '{headerstr}' of file '{path}'.")
            idx += 1

            for key in [log.KEY_BEST_F, log.KEY_LAST_IMPROVEMENT_FE,
                        log.KEY_LAST_IMPROVEMENT_TIME_MILLIS,
                        log.KEY_TOTAL_FES, log.KEY_TOTAL_TIME_MILLIS]:
                if csv(key) != header[idx:(idx + CSV_COLS)]:
                    raise ValueError(
                        f"Expected to find '{key}.*' keys from index "
                        f"{idx} on in header "
                        f"'{headerstr}' of file '{path}', expected "
                        f"{csv(key)} but got {header[idx:(idx + CSV_COLS)]}.")
                idx += CSV_COLS

            has_goal_f: int = 0
            has_best_f_scaled: bool = False
            has_n_success: bool = False
            has_success_fes: bool = False
            has_success_time: bool = False
            has_ert_fes: bool = False
            has_ert_time: bool = False
            has_max_fes: int = 0
            has_max_time: int = 0
            while idx <= len(header):
                if header[idx] == log.KEY_GOAL_F:
                    has_goal_f = 1
                    idx += 1
                elif header[idx].startswith(log.KEY_GOAL_F):
                    has_goal_f = 2
                    if csv(log.KEY_GOAL_F) != header[idx:(idx + CSV_COLS)]:
                        raise ValueError(
                            f"Expected to find '{log.KEY_GOAL_F}.*' keys from "
                            f"index {idx} on in header "
                            f"'{headerstr}' of file '{path}'.")
                    idx += CSV_COLS

                if idx >= len(header):
                    break

                if header[idx].startswith(KEY_BEST_F_SCALED):
                    has_best_f_scaled = True
                    if csv(KEY_BEST_F_SCALED) != \
                            header[idx:(idx + CSV_COLS)]:
                        raise ValueError(
                            f"Expected to find '{KEY_BEST_F_SCALED}.*' "
                            f"keys from index {idx} on in header "
                            f"'{headerstr}' of file '{path}'.")
                    idx += CSV_COLS

                if idx >= len(header):
                    break

                if header[idx] == KEY_N_SUCCESS:
                    has_n_success = True
                    idx += 1

                if idx >= len(header):
                    break

                if header[idx].startswith(KEY_SUCCESS_FES):
                    has_success_fes = True
                    if csv(KEY_SUCCESS_FES) != header[idx:(idx + CSV_COLS)]:
                        raise ValueError(
                            f"Expected to find '{KEY_SUCCESS_FES}.*' "
                            f"keys from index {idx} on in header "
                            f"'{headerstr}' of file '{path}'.")
                    idx += CSV_COLS

                if idx >= len(header):
                    break

                if header[idx].startswith(KEY_SUCCESS_TIME_MILLIS):
                    has_success_time = True
                    if csv(KEY_SUCCESS_TIME_MILLIS) != \
                            header[idx:(idx + CSV_COLS)]:
                        raise ValueError(
                            f"Expected to find '{KEY_SUCCESS_TIME_MILLIS}.*' "
                            f"keys from index {idx} on in header "
                            f"'{headerstr}' of file '{path}'.")
                    idx += CSV_COLS

                if idx >= len(header):
                    break

                if header[idx] == KEY_ERT_FES:
                    has_ert_fes = True
                    idx += 1

                if idx >= len(header):
                    break

                if header[idx] == KEY_ERT_TIME_MILLIS:
                    has_ert_time = True
                    idx += 1

                if idx >= len(header):
                    break

                if header[idx] == log.KEY_MAX_FES:
                    has_max_fes = 1
                    idx += 1
                elif header[idx].startswith(log.KEY_MAX_FES):
                    has_max_fes = 2
                    if csv(log.KEY_MAX_FES) != header[idx:(idx + CSV_COLS)]:
                        raise ValueError(
                            f"Expected to find '{log.KEY_MAX_FES}.*' keys"
                            f" from index {idx} on in header "
                            f"'{headerstr}' of file '{path}'.")
                    idx += CSV_COLS

                if idx >= len(header):
                    break

                if header[idx] == log.KEY_MAX_TIME_MILLIS:
                    has_max_time = 1
                    idx += 1
                elif header[idx].startswith(log.KEY_MAX_TIME_MILLIS):
                    has_max_time = 2
                    if csv(log.KEY_MAX_FES) != header[idx:(idx + CSV_COLS)]:
                        raise ValueError(
                            f"Expected to find '{log.KEY_MAX_TIME_MILLIS}.*' "
                            f"keys from index {idx} on in header "
                            f"'{headerstr}' of file '{path}'.")
                    idx += CSV_COLS

                break

            if len(header) > idx:
                raise ValueError(
                    f"Unexpected item '{header[idx]}' in header "
                    f"'{header}' of file '{path}'.")

            while True:
                lines = rd.readlines(100)
                if (lines is None) or (len(lines) <= 0):
                    break
                for line in lines:
                    row = [ss.strip() for ss in line.strip().split(sep)]

                    idx = 0
                    algo: str | None = None
                    inst: str | None = None
                    n: int
                    goal_f: None | int | float | Statistics = None
                    best_f_scaled: Statistics | None = None
                    n_success: int | None = None
                    success_fes: Statistics | None = None
                    success_time: Statistics | None = None
                    ert_fes: int | float | None = None
                    ert_time: int | float | None = None
                    max_fes: int | Statistics | None = None
                    max_time: int | Statistics | None = None

                    try:
                        if has_algorithm:
                            algo = sanitize_name(row[0])
                            idx += 1

                        if has_instance:
                            inst = sanitize_name(row[idx])
                            idx += 1

                        n = int(row[idx])
                        idx += 1

                        best_f: Statistics = Statistics.from_csv(
                            n, row[idx:(idx + CSV_COLS)])
                        idx += CSV_COLS

                        last_improv_fe: Statistics = Statistics.from_csv(
                            n, row[idx:(idx + CSV_COLS)])
                        idx += CSV_COLS

                        last_improv_time: Statistics = Statistics.from_csv(
                            n, row[idx:(idx + CSV_COLS)])
                        idx += CSV_COLS

                        total_fes: Statistics = Statistics.from_csv(
                            n, row[idx:(idx + CSV_COLS)])
                        idx += CSV_COLS

                        total_time: Statistics = Statistics.from_csv(
                            n, row[idx:(idx + CSV_COLS)])
                        idx += CSV_COLS

                        if has_goal_f == 1:
                            goal_f = str_to_intfloat(row[idx])
                            idx += 1
                        elif has_goal_f == 2:
                            goal_f = Statistics.from_csv(
                                n, row[idx:(idx + CSV_COLS)])
                            idx += CSV_COLS

                        if has_best_f_scaled:
                            best_f_scaled = Statistics.from_csv(
                                n, row[idx:(idx + CSV_COLS)])
                            idx += CSV_COLS

                        if has_n_success:
                            n_success = int(row[idx])
                            idx += 1

                        if has_success_fes:
                            if n_success > 0:
                                success_fes = Statistics.from_csv(
                                    n_success, row[idx:(idx + CSV_COLS)])
                            else:
                                success_fes = None
                            idx += CSV_COLS

                        if has_success_time:
                            if n_success > 0:
                                success_time = Statistics.from_csv(
                                    n_success, row[idx:(idx + CSV_COLS)])
                            else:
                                success_time = None
                            idx += CSV_COLS

                        if has_ert_fes:
                            if n_success > 0:
                                ert_fes = str_to_intfloat(row[idx])
                            else:
                                ert_fes = inf
                            idx += 1

                        if has_ert_time:
                            if n_success > 0:
                                ert_time = str_to_intfloat(row[idx])
                            else:
                                ert_time = inf
                            idx += 1

                        if has_max_fes == 1:
                            max_fes = int(row[idx])
                            idx += 1
                        elif has_max_fes == 2:
                            max_fes = Statistics.from_csv(
                                n, row[idx:(idx + CSV_COLS)])
                            idx += CSV_COLS

                        if has_max_time == 1:
                            max_time = int(row[idx])
                            idx += 1
                        elif has_max_time == 2:
                            max_time = Statistics.from_csv(
                                n, row[idx:(idx + CSV_COLS)])
                            idx += CSV_COLS

                    except BaseException as be:
                        raise ValueError(
                            f"Invalid row '{line}' in file '{path}'.") from be

                    if len(row) != idx:
                        raise ValueError("Invalid number of columns in row "
                                         f"'{line}' in file '{path}'.")
                    consumer(EndStatistics(
                        algo, inst, n, best_f, last_improv_fe,
                        last_improv_time, total_fes, total_time, goal_f,
                        best_f_scaled, n_success, success_fes, success_time,
                        ert_fes, ert_time, max_fes, max_time))

        logger("Finished reading end result statistics from CSV "
               f"file '{path}'.")

    @staticmethod
    def getter(dimension: str) -> Callable[["EndStatistics"],
                                           int | float | None]:
        """
        Create a function that obtains the given dimension from EndStatistics.

        :param dimension: the dimension
        :returns: a callable that returns the value corresponding to the
            dimension
        """
        if not isinstance(dimension, str):
            raise type_error(dimension, "dimension", str)
        if dimension in _GETTERS_0:
            return _GETTERS_0[dimension]

        ssi: int = dimension.find(SCOPE_SEPARATOR)
        if ssi <= 0:
            raise ValueError(f"unknown dimension '{dimension}'.")
        scope: str = dimension[:ssi]
        dim: str = dimension[ssi + 1:]
        if (len(scope) <= 0) or (len(dim) <= 0):
            raise ValueError(
                f"invalid dimension '{dimension}', has "
                f"scope '{scope}' and sub-dimension '{dim}'")

        if scope not in _GETTERS_1:
            raise ValueError(
                f"invalid dimension '{dimension}', has "
                f"unknown scope '{scope}' and sub-dimension '{dim}'")

        l1 = _GETTERS_1[scope]

        if dim != KEY_STDDEV:
            l2 = Statistics.getter(dim)

            def __inner_sat(s: EndStatistics, ll1=l1, ll2=l2) \
                    -> int | float | None:
                a: Final[Statistics] = ll1(s)
                if a is None:
                    return None
                if isinstance(a, (int, float)):
                    return a  # max, min, med, mean = a
                return ll2(a)  # apply statistics getter
            return __inner_sat

        def __inner_sd(s: EndStatistics, ll1=l1) -> int | float | None:
            a: Final[Statistics] = ll1(s)
            if a is None:
                return None
            if isinstance(a, (int, float)):
                return 0  # sd of a single number = 0
            return a.stddev
        return __inner_sd


# Run end-results to stat file if executed as script
if __name__ == "__main__":
    help_screen(
        "build end results statistics-CSV file", __file__,
        "Convert log files or the end-results CSV obtained "  # nosem
        "with moptipy to the end results statistics CSV format.",  # nosem
        [("source",
          "either the directory of either moptipy log files or "  # nosem
          "the path to the end-results CSV file"),  # nosem
         ("dest_file",
          "the path to which we want to write the end "  # nosem
          "statistics CSV file.")])  # nosem
    if len(sys.argv) != 3:
        raise ValueError("two command line arguments expected")

    src_path = Path.path(sys.argv[1])
    end_results: Final[list[EndResult]] = []
    if src_path.is_file():
        logger(f"'{src_path}' identifies file, load as end-results csv")
        EndResult.from_csv(src_path, end_results.append)
    else:
        logger(f"'{src_path}' identifies directory, load it as log files")
        EndResult.from_logs(src_path, end_results.append)

    end_stats: Final[list[EndStatistics]] = []
    EndStatistics.from_end_results(end_results, end_stats.append)
    EndStatistics.to_csv(end_stats, sys.argv[2])



"""Print a help screen."""

import os.path
import sys
from typing import Final, Iterable

from moptipy.utils.console import logger
from moptipy.utils.path import _canonicalize_path
from moptipy.utils.types import type_error

#: The python interpreter in long form.
__INTERPRETER_LONG: Final[str] = _canonicalize_path(sys.executable)


def __get_python_interpreter_short() -> str:
    """
    Get the python interpreter.

    :returns: the fully-qualified path
    """
    bn = os.path.basename(__INTERPRETER_LONG)
    if bn.startswith("python3."):
        bn2 = bn[:7]
        interp2 = os.path.join(os.path.dirname(__INTERPRETER_LONG), bn2)
        if os.path.exists(interp2) and os.path.isfile(interp2):
            if _canonicalize_path(interp2) == __INTERPRETER_LONG:
                return bn2
    return bn


#: The python interpreter in short form.
__INTERPRETER_SHORT: Final[str] = __get_python_interpreter_short()
del __get_python_interpreter_short

#: the length of the base path
__BASE_PATH: Final[str] = _canonicalize_path(os.path.dirname(
    os.path.dirname(os.path.dirname(_canonicalize_path(__file__))))) + os.sep


def help_screen(title: str, file: str, text: str = "",
                args: Iterable[tuple[str, str]
                               | tuple[str, str, bool]] = ()) -> None:
    """
    Print a help screen on the console.

    :param title: the program title
    :param file: the calling python script
    :param text: the text explaining the program
    :param args: the command line arguments, as sequence of tuples
        `(title, description)` or `(title, description, optional)`
    :return: nothing
    """
    if not isinstance(title, str):
        raise type_error(title, "title", str)
    if not isinstance(file, str):
        raise type_error(file, "file", str)
    if not isinstance(text, str):
        raise type_error(text, "text", str)
    if not isinstance(args, Iterable):
        raise type_error(args, "args", Iterable)

    # get the module
    module: str = _canonicalize_path(file)
    end: int = len(module)
    start: int = 0
    if module.endswith(".py"):
        end -= 3
    if module.startswith(__BASE_PATH):
        start += len(__BASE_PATH)
    module = module[start:end].replace(os.sep, ".")
    mid: Final[str] = "' '"
    sargs: Final[str] = " ".join(
        f"[{t[0]}]" if (len(t) > 2) and t[2]  # type: ignore
        else t[0] for t in args)
    # prepare the text dump
    cons: Final[list[str]] = [
        title,
        f"usage: {__INTERPRETER_SHORT} -m {module} {sargs}",
        f" call: {__INTERPRETER_LONG} -m {module} '{mid.join(sys.argv[1:])}'"]

    if len(text) > 0:
        cons.append(text)

    # iterate over the arguments
    for arg in args:
        if not isinstance(arg, tuple):
            raise type_error(arg, "args[i]", tuple)
        if not 1 < len(arg) < 4:
            raise ValueError(
                f"invalid argument tuple {arg}, should have length 2 or 3.")
        if not isinstance(arg[0], str):
            raise type_error(arg[0], "arg[0]", str)
        if len(arg[0]) <= 0:
            raise ValueError(f"invalid arg[0]: '{arg[0]}'")
        if not isinstance(arg[1], str):
            raise type_error(arg[1], "arg[1]", str)
        if len(arg[0]) <= 0:
            raise ValueError(f"invalid arg[1]: '{arg[1]}'")
        s = f"- {arg[0]}: {arg[1]}"
        if len(arg) > 2:
            if not isinstance(arg[2], bool):  # type: ignore
                raise type_error(arg[2],  # type: ignore
                                 "arg[2]", bool)
            if arg[2]:  # type: ignore
                s = f"{s} [optional]"
        cons.append(s)
    logger("\n".join(cons))



"""A section consumer can load the sections from a log file."""

from os import listdir
from os.path import basename, dirname, isdir, isfile, join
from typing import Final

from moptipy.api import logging
from moptipy.utils.console import logger
from moptipy.utils.logger import COMMENT_CHAR, SECTION_END, SECTION_START
from moptipy.utils.nputils import rand_seed_check
from moptipy.utils.path import Path
from moptipy.utils.strings import PART_SEPARATOR, sanitize_name


class LogParser:
    """
    A log parser can parse a log file and separate the sections.

    The log parser is designed to load data from text files generated
    by :class:`~moptipy.utils.logger.FileLogger`. It can also recursively
    parse directories.
    """

    def __init__(self,
                 print_begin_end: bool = True,
                 print_file_start: bool = False,
                 print_file_end: bool = False,
                 print_dir_start: bool = True,
                 print_dir_end: bool = True):
        """
        Initialize the log parser.

        :param print_begin_end: log to stdout when starting and ending
            a recursive directory parsing process
        :param print_file_start: log to stdout when opening a file
        :param print_file_end: log to stdout when closing a file
        :param print_dir_start: log to stdout when entering a directory
        :param print_dir_end: log to stdout when leaving a directory
        """
        self.__print_begin_end: Final[bool] = print_begin_end
        self.__print_file_start: Final[bool] = print_file_start
        self.__print_file_end: Final[bool] = print_file_end
        self.__print_dir_start: Final[bool] = print_dir_start
        self.__print_dir_end: Final[bool] = print_dir_end
        self.__depth: int = 0

    # noinspection PyUnusedLocal
    # noinspection PyMethodMayBeStatic
    def start_dir(self, path: Path) -> bool:
        """
        Enter a directory to parse all files inside.

        This method is called by
        :meth:`~moptipy.evaluation.log_parser.LogParser.parse_dir`.
        If it returns `True`, every sub-directory inside of it will be passed
        to :meth:`~moptipy.evaluation.log_parser.LogParser.start_dir`
        and every file will be passed to
        :meth:`~moptipy.evaluation.log_parser.LogParser.start_file`.
        Only if `True` is returned,
        :meth:`~moptipy.evaluation.log_parser.LogParser.end_dir` will be
        invoked and its return value will be the return value of
        :meth:`~moptipy.evaluation.log_parser.LogParser.parse_dir`. If `False`
        is returned, then
        :meth:`~moptipy.evaluation.log_parser.LogParser.parse_dir` will return
        immediately and return `True`.

        :param path: the path of the directory
        :return: `True` if all the files and sub-directories inside the
            directory should be processed, `False` if this directory should
            be skipped and parsing should continue with the next sibling
            directory
        """
        del path
        return True

    # noinspection PyUnusedLocal
    # noinspection PyMethodMayBeStatic
    def end_dir(self, path: Path) -> bool:
        """
        Enter a directory to parse all files inside.

        This method is called by
        :meth:`~moptipy.evaluation.log_parser.LogParser.parse_dir`.
        If it returns `True`, every sub-directory inside of it will be passed
        to :meth:`~moptipy.evaluation.log_parser.LogParser.start_dir`
        and every file will be passed to
        :meth:`~moptipy.evaluation.log_parser.LogParser.start_file`.

        :param path: the path of the directory
        :return: `True` if all the files and sub-directories inside the
            directory should be processed, `False` if this directory should
            be skipped and parsing should continue with the next sibling
            directory
        """
        del path
        return True

    # noinspection PyUnusedLocal
    # noinspection PyMethodMayBeStatic
    def start_file(self, path: Path) -> bool:
        """
        Decide whether to start parsing a file.

        This method is called by
        :meth:`~moptipy.evaluation.log_parser.LogParser.parse_file`. If it
        returns `True`, then we will open and parse the file. If it returns
        `False`, then the fill will not be parsed and
        :meth:`~moptipy.evaluation.log_parser.LogParser.parse_file` will
        return `True` immediately.

        :param path: the file path
        :return: `True` if the file should be parsed, `False` if it should be
            skipped (and
            :meth:`~moptipy.evaluation.log_parser.LogParser.parse_file` should
            return `True`).
        """
        return path.endswith(logging.FILE_SUFFIX)

    # noinspection PyMethodMayBeStatic
    def start_section(self, title: str) -> bool:
        """
        Start a section.

        If this method returns `True`, then all the lines of text of the
        section `title` will be read and together passed to
        :meth:`~moptipy.evaluation.log_parser.LogParser.lines`.
        If this method returns `False`, then the section will be skipped
        and we fast-forward to the next section, if any, or to the call
        of :meth:`~moptipy.evaluation.log_parser.LogParser.end_file`.

        :param title: the section title
        :return: `True` if the section data should be loaded and passed to
            :meth:`lines`, `False` of the section can be skipped. In that
            case, we will fast-forward to the next
            :meth:`~moptipy.evaluation.log_parser.LogParser.start_section`.
        """
        if not title:
            raise ValueError(f"Title cannot be empty, but is '{title}'.")
        if title.startswith(logging.ERROR_SECTION_PREFIX):
            raise ValueError(f"Encountered error section '{title}'.")
        return False

    # noinspection PyUnusedLocal
    # noinspection PyMethodMayBeStatic
    def lines(self, lines: list[str]) -> bool:
        """
        Consume all the lines from a section.

        This method receives the complete text of a section, where all lines
        are separated and put into one list `lines`. Each line is stripped
        from whitespace and comments, empty lines are omitted.
        If this method returns `True`, we will continue parsing the file and
        move to the next section, if any, or directly to the end of the file
        parsing process (and call
        :meth:`~moptipy.evaluation.log_parser.LogParser.end_file`) if there is
        no more section in the file.

        :param lines: the lines to consume
        :return: `True` if further parsing is necessary and the next section
            should be fed to
            :meth:`~moptipy.evaluation.log_parser.LogParser.start_section`,
            `False` if the parsing
            process can be terminated, in which case we will fast-forward to
            :meth:`~moptipy.evaluation.log_parser.LogParser.end_file`
        """
        del lines
        return True

    # noinspection PyMethodMayBeStatic
    def end_file(self) -> bool:
        """
        End a file.

        This method is invoked when we have reached the end of the current
        file. Its return value, `True` or `False`, will then be returned
        by :meth:`~moptipy.evaluation.log_parser.LogParser.parse_file`, which
        is the entry point for the file parsing process.

        :return: the return value to be returned by
            :meth:`~moptipy.evaluation.log_parser.LogParser.parse_file`
        """
        return True

    def parse_file(self, path: str) -> bool:
        """
        Parse the contents of a file.

        This method first calls the function
        :meth:`~moptipy.evaluation.log_parser.LogParser.start_file` to see
        whether the file should be parsed. If
        :meth:`~moptipy.evaluation.log_parser.LogParser.start_file` returns
        `True`, then the file is parsed. If
        :meth:`~moptipy.evaluation.log_parser.LogParser.start_file` returns
        `False`, then this method returns `False` directly.
        If the file is parsed, then
        :meth:`~moptipy.evaluation.log_parser.LogParser.start_section` will
        be invoked for each section (until the parsing is finished) and
        :meth:`~moptipy.evaluation.log_parser.LogParser.lines` for
        each section content (if requested). At the end,
        :meth:`~moptipy.evaluation.log_parser.LogParser.end_file` is invoked.

        This method can either be called directly or is called by
        :meth:`~moptipy.evaluation.log_parser.LogParser.parse_dir`. In the
        latter case, if
        :meth:`~moptipy.evaluation.log_parser.LogParser.parse_file` returned
        `True`, the next file in the current directory will be parsed. If it
        returns `False`, then no file located in the current directory will be
        parsed, while other directories and/or sub-directories will still be
        processed.

        :param path: the file to parse
        :return: the return value received from invoking
            :meth:`~moptipy.evaluation.log_parser.LogParser.end_file`
        """
        file: Final[Path] = Path.file(path)

        retval: bool
        try:
            retval = self.start_file(file)
        except BaseException as be:
            raise ValueError(f"Error when starting file '{file}'") from be

        if retval:
            if self.__print_file_start:
                logger(f"parsing file '{file}'.")
        else:
            if self.__print_file_start:
                logger(f"skipping file '{file}'.")
            return True

        lines: list[str] = []
        buffer: list[str] = []
        state: int = 0
        wants_section: bool = False
        sec_end: str = ""
        section: str = ""
        sect_start: Final[str] = SECTION_START
        sect_end: Final[str] = SECTION_END
        cmt_chr: Final[str] = COMMENT_CHAR

        index: int = 0
        with file.open_for_read() as handle:
            while True:

                # get the next line
                if index >= len(buffer):
                    try:
                        buffer = handle.readlines(128)
                    except BaseException as be:
                        raise ValueError(
                            f"Error when reading lines from file '{file}' "
                            f"while in section '{section}'."
                            if state == 1 else
                            f"Error when reading lines from file '{file}'.") \
                            from be
                    if (buffer is None) or (len(buffer) <= 0):
                        break
                    index = 0

                orig_cur = buffer[index]
                index += 1

                # strip next line from comments and white space
                cur = orig_cur.strip()
                if len(cur) <= 0:
                    continue

                i = cur.find(cmt_chr)
                if i >= 0:
                    cur = cur[:i].strip()
                    if len(cur) <= 0:
                        continue

                if state in (0, 2):
                    if not cur.startswith(sect_start):
                        ValueError("Line should start with "
                                   f"'{sect_start}' but is "
                                   f"'{orig_cur}' in file '{file}'.")
                    section = cur[len(sect_start):]
                    if len(section) <= 0:
                        ValueError("Section title cannot be empty in "
                                   f"'{file}', but encountered '{orig_cur}'.")
                    state = 1
                    sec_end = sect_end + section
                    wants_section = self.start_section(section)
                elif state == 1:
                    if cur == sec_end:
                        state = 2
                        if wants_section:
                            try:
                                do_next = self.lines(lines)
                            except BaseException as be:
                                raise ValueError(
                                    "Error when processing section "
                                    f"'{section}' in file '{file}'.") from be
                            lines.clear()
                            if not do_next:
                                break
                    elif wants_section:
                        lines.append(cur)

        if state == 0:
            raise ValueError(f"Log file '{file}' contains no section.")
        if state == 1:
            raise ValueError(f"Log file '{file}' ended before"
                             f"encountering '{sec_end}'.")

        try:
            retval = self.end_file()
        except BaseException as be:
            raise ValueError("Error when ending section parsing "
                             f"of file '{file}'.") from be
        if self.__print_file_end:
            logger(f"finished parsing file '{file}'.")
        return retval

    def parse_dir(self, path: str) -> bool:
        """
        Recursively parse the given directory.

        :param path: the path to the directory
        :return: `True` either if
            :meth:`~moptipy.evaluation.log_parser.LogParser.start_dir`
            returned `False` or
            :meth:`~moptipy.evaluation.log_parser.LogParser.end_dir` returned
            `True`, `False` otherwise
        """
        folder: Final[Path] = Path.directory(path)

        if self.__depth <= 0:
            if self.__depth == 0:
                if self.__print_begin_end:
                    logger("beginning recursive parsing of "
                           f"directory '{folder}'.")
            else:
                raise ValueError(
                    f"Depth must be >= 0, but is {self.__depth}??")
        self.__depth += 1

        if not self.start_dir(folder):
            if self.__print_dir_start:
                logger(f"skipping directory '{folder}'.")
            return True
        if self.__print_dir_start:
            logger(f"entering directory '{folder}'.")

        do_files = True
        do_dirs = True
        for sub in listdir(folder):
            sub = Path.path(join(folder, sub))
            if isfile(sub):
                if do_files:
                    if not self.parse_file(sub):
                        logger(f"will parse no more files in '{folder}'.")
                        if not do_dirs:
                            break
                        do_files = False
            elif isdir(sub):
                if do_dirs:
                    if not self.parse_dir(sub):
                        logger("will parse no more sub-directories "
                               f"of '{folder}'.")
                        if not do_files:
                            break
                        do_dirs = False

        retval = self.end_dir(folder)
        if self.__print_dir_end:
            logger(f"finished parsing directory '{folder}'.")

        self.__depth -= 1
        if self.__depth <= 0:
            if self.__depth == 0:
                if self.__print_begin_end:
                    logger(f"finished recursive parsing of "
                           f"directory '{folder}'.")
            else:
                raise ValueError(
                    f"Depth must be >= 0, but is {self.__depth}??")

        return retval

    def parse(self, path: str) -> bool:
        """
        Parse either a directory or a file.

        If `path` identifies a file,
        :meth:`~moptipy.evaluation.log_parser.LogParser.parse_file` is invoked
        and its result is returned. If `path` identifies a directory, then
        :meth:`~moptipy.evaluation.log_parser.LogParser.parse_dir` is invoked
        and its result is returned.

        :param path: a path identifying either a directory or a file.
        :return: the result of the appropriate parsing routing
        :raises ValueError: if `path` does not identify a directory or file
        """
        npath = Path.path(path)
        if isfile(npath):
            return self.parse_file(npath)
        if isdir(npath):
            return self.parse_dir(npath)
        raise ValueError(
            f"Path {npath} is neither a file nor a directory?")


class ExperimentParser(LogParser):
    """A log parser following our pre-defined experiment structure."""

    def __init__(self):
        """Initialize the experiment parser."""
        super().__init__(print_begin_end=True, print_dir_start=True)

        #: The name of the algorithm to which the current log file belongs.
        self.algorithm: str | None = None

        #: The name of the instance to which the current log file belongs.
        self.instance: str | None = None

        #: The random seed of the current log file.
        self.rand_seed: int | None = None

    def start_file(self, path: Path) -> bool:
        """
        Decide whether to start parsing a file and setup meta-data.

        :param path: the file path
        :return: `True` if the file should be parsed, `False` if it should be
            skipped (and
            :meth:`~moptipy.evaluation.log_parser.LogParser.parse_file` should
            return `True`).
        """
        if not super().start_file(path):
            return False

        inst_dir = dirname(path)
        algo_dir = dirname(inst_dir)
        self.instance = sanitize_name(basename(inst_dir))
        self.algorithm = sanitize_name(basename(algo_dir))

        start = f"{self.algorithm}{PART_SEPARATOR}" \
                f"{self.instance}{PART_SEPARATOR}0x"
        base = basename(path)
        if not base.startswith(start):
            raise ValueError(
                f"File name of '{path}' should start with '{start}'.")
        self.rand_seed = rand_seed_check(int(
            base[len(start):(-len(logging.FILE_SUFFIX))], base=16))

        return True

    def end_file(self) -> bool:
        """Finalize parsing a file."""
        self.rand_seed = None
        self.algorithm = None
        self.instance = None
        return super().end_file()



"""
In this package, we provide implementations of metaheuristic algorithms.

We divide them into single-objective algorithms, which are provided in the
package :mod:`moptipy.algorithms.so`, and multi-objective algorithms,
which can be found in package :mod:`moptipy.algorithms.mo`.

Methods which are based on random sampling and random walks are unaffected by
the number of objective functions and therefore can be found in this package
directly.

The black-box methods are given directly in these package, more specialized
algorithms will be placed in sub-packages corresponding to their requirements
for the search- and solution space.
"""



"""Approximate the ECDF over the ERT to reach certain goals."""

from dataclasses import dataclass

from moptipy.evaluation.ecdf import Ecdf
from moptipy.evaluation.ert import compute_single_ert
from moptipy.evaluation.progress import Progress


@dataclass(frozen=True, init=False, order=True)
class ErtEcdf(Ecdf):
    """The ERT-ECDF."""

    def time_label(self) -> str:
        """
        Get the time axis label.

        :return: the time key
        """
        return f"ERT\u2009[{self.time_unit}]"

    def _time_key(self) -> str:
        """
        Get the time key.

        :return: the time key
        """
        return f"ert[{super()._time_key()}]"

    @staticmethod
    def _compute_times(source: list[Progress],
                       goal: int | float) -> list[float]:
        """
        Compute the times for the given goals.

        :param source: the source array
        :param goal: the goal value
        :return: a list of times
        """
        return [compute_single_ert(source, goal)]

    # noinspection PyUnusedLocal
    @staticmethod
    def _get_div(n: int, n_insts: int) -> int:
        """
        Get the divisor.

        :param n: the number of runs
        :param n_insts: the number of instances
        :return: the divisor
        """
        del n
        return n_insts



"""
Numerical optimization algorithms working on real-valued vectors.

The search space here is always
:class:`~moptipy.spaces.vectorspace.VectorSpace`, which prescribes a box in
the space of `n`-dimensional real vectors that is constraint by an upper and a
lower bound in each dimension.
Algorithms working on such spaces can use the rich mathematical relationships
and operations defined over the continuous spaces.
"""



"""Plot a set of ERT objects into one figure."""
from typing import Any, Callable, Final, Iterable, cast

import numpy as np
from matplotlib.artist import Artist  # type: ignore
from matplotlib.axes import Axes  # type: ignore
from matplotlib.figure import Figure, SubplotBase  # type: ignore

import moptipy.utils.plot_defaults as pd
import moptipy.utils.plot_utils as pu
from moptipy.evaluation.axis_ranger import AxisRanger
from moptipy.evaluation.base import get_algorithm, get_instance, sort_key
from moptipy.evaluation.ert import Ert
from moptipy.evaluation.styler import Styler
from moptipy.utils.lang import Lang
from moptipy.utils.types import type_error


def plot_ert(erts: Iterable[Ert],
             figure: SubplotBase | Figure,
             x_axis: AxisRanger | Callable[[str], AxisRanger]
             = AxisRanger.for_axis,
             y_axis: AxisRanger | Callable[[str], AxisRanger]
             = AxisRanger.for_axis,
             legend: bool = True,
             distinct_colors_func: Callable[[int], Any] =
             pd.distinct_colors,
             distinct_line_dashes_func: Callable[[int], Any] =
             pd.distinct_line_dashes,
             importance_to_line_width_func: Callable[[int], float] =
             pd.importance_to_line_width,
             importance_to_alpha_func: Callable[[int], float] =
             pd.importance_to_alpha,
             importance_to_font_size_func: Callable[[int], float] =
             pd.importance_to_font_size,
             x_grid: bool = True,
             y_grid: bool = True,
             x_label: None | str | Callable[[str], str] = Lang.translate,
             x_label_inside: bool = True,
             y_label: None | str | Callable[[str], str] =
             Lang.translate_func("ERT"),
             y_label_inside: bool = True,
             inst_priority: float = 0.666,
             algo_priority: float = 0.333) -> Axes:
    """
    Plot a set of Ert functions into one chart.

    :param erts: the iterable of Ert functions
    :param figure: the figure to plot in
    :param x_axis: the x_axis ranger
    :param y_axis: the y_axis ranger
    :param legend: should we plot the legend?
    :param distinct_colors_func: the function returning the palette
    :param distinct_line_dashes_func: the function returning the line styles
    :param importance_to_line_width_func: the function converting importance
        values to line widths
    :param importance_to_alpha_func: the function converting importance
        values to alphas
    :param importance_to_font_size_func: the function converting importance
        values to font sizes
    :param x_grid: should we have a grid along the x-axis?
    :param y_grid: should we have a grid along the y-axis?
    :param x_label: a callable returning the label for the x-axis, a label
        string, or `None` if no label should be put
    :param x_label_inside: put the x-axis label inside the plot (so that
        it does not consume additional vertical space)
    :param y_label: a callable returning the label for the y-axis, a label
        string, or `None` if no label should be put
    :param y_label_inside: put the y-axis label inside the plot (so that
        it does not consume additional horizontal space)
    :param inst_priority: the style priority for instances
    :param algo_priority: the style priority for algorithms
    :returns: the axes object to allow you to add further plot elements
    """
    # Before doing anything, let's do some type checking on the parameters.
    # I want to ensure that this function is called correctly before we begin
    # to actually process the data. It is better to fail early than to deliver
    # some incorrect results.
    if not isinstance(erts, Iterable):
        raise type_error(erts, "erts", Iterable)
    if not isinstance(figure, (SubplotBase, Figure)):
        raise type_error(figure, "figure", (SubplotBase, Figure))
    if not isinstance(legend, bool):
        raise type_error(legend, "legend", bool)
    if not callable(distinct_colors_func):
        raise type_error(
            distinct_colors_func, "distinct_colors_func", call=True)
    if not callable(distinct_line_dashes_func):
        raise type_error(
            distinct_line_dashes_func, "distinct_line_dashes_func", call=True)
    if not callable(distinct_line_dashes_func):
        raise type_error(importance_to_line_width_func,
                         "importance_to_line_width_func", call=True)
    if not callable(importance_to_alpha_func):
        raise type_error(
            importance_to_alpha_func, "importance_to_alpha_func", call=True)
    if not callable(importance_to_font_size_func):
        raise type_error(importance_to_font_size_func,
                         "importance_to_font_size_func", call=True)
    if not isinstance(x_grid, bool):
        raise type_error(x_grid, "x_grid", bool)
    if not isinstance(y_grid, bool):
        raise type_error(y_grid, "y_grid", bool)
    if not ((x_label is None) or callable(x_label)
            or isinstance(x_label, str)):
        raise type_error(x_label, "x_label", (str, None), call=True)
    if not isinstance(x_label_inside, bool):
        raise type_error(x_label_inside, "x_label_inside", bool)
    if not ((y_label is None) or callable(y_label)
            or isinstance(y_label, str)):
        raise type_error(y_label, "y_label", (str, None), call=True)
    if not isinstance(y_label_inside, bool):
        raise type_error(y_label_inside, "y_label_inside", bool)
    if not isinstance(inst_priority, float):
        raise type_error(inst_priority, "inst_priority", float)
    if not isinstance(algo_priority, float):
        raise type_error(algo_priority, "algo_priority", float)

    # First, we try to find groups of data to plot together in the same
    # color/style.
    instances: Final[Styler] = Styler(
        key_func=get_instance,
        none_name=Lang.translate("all_insts"),
        priority=inst_priority)
    algorithms: Final[Styler] = Styler(
        key_func=get_algorithm,
        none_name=Lang.translate("all_algos"),
        priority=algo_priority)
    x_dim: str | None = None
    y_dim: str | None = None
    source: list[Ert] = cast(list[Ert], erts) if isinstance(erts, list) \
        else list(erts)
    del erts

    # First pass: find out the instances and algorithms
    for ert in source:
        if not isinstance(ert, Ert):
            raise type_error(ert, "ert data source", Ert)
        instances.add(ert)
        algorithms.add(ert)

        # Validate that we have consistent time and objective units.
        if x_dim is None:
            x_dim = ert.f_name
        elif x_dim != ert.f_name:
            raise ValueError(
                f"F-units {x_dim} and {ert.f_name} do not fit!")

        if y_dim is None:
            y_dim = ert.time_unit
        elif y_dim != ert.time_unit:
            raise ValueError(
                f"Time units {y_dim} and {ert.time_unit} do not fit!")

    if (x_dim is None) or (y_dim is None) or (len(source) <= 0):
        raise ValueError("Illegal state?")

    source.sort(key=sort_key)

    def __set_importance(st: Styler) -> None:
        none = 1
        not_none = 0
        none_lw = importance_to_line_width_func(none)
        not_none_lw = importance_to_line_width_func(not_none)
        st.set_line_width(lambda p: [none_lw if i <= 0 else not_none_lw
                                     for i in range(p)])
        none_a = importance_to_alpha_func(none)
        not_none_a = importance_to_alpha_func(not_none)
        st.set_line_alpha(lambda p: [none_a if i <= 0 else not_none_a
                                     for i in range(p)])

    # determine the style groups
    groups: list[Styler] = []
    instances.finalize()
    algorithms.finalize()

    if instances.count > 1:
        groups.append(instances)
    if algorithms.count > 1:
        groups.append(algorithms)

    if len(groups) > 0:
        groups.sort()
        groups[0].set_line_color(distinct_colors_func)

        if len(groups) > 1:
            groups[1].set_line_dash(distinct_line_dashes_func)

    # If we only have <= 2 groups, we can mark None and not-None values with
    # different importance.
    if instances.has_none and (instances.count > 1):
        __set_importance(instances)
    elif algorithms.has_none and (algorithms.count > 1):
        __set_importance(algorithms)

    # we will collect all lines to plot in plot_list
    plot_list: list[dict] = []

    # set up the axis rangers
    if callable(x_axis):
        x_axis = x_axis(x_dim)
    if not isinstance(x_axis, AxisRanger):
        raise type_error(x_axis, "x_axis", AxisRanger)

    if callable(y_axis):
        y_axis = y_axis(y_dim)
    if not isinstance(y_axis, AxisRanger):
        raise type_error(y_axis, "y_axis", AxisRanger)

    # first we collect all progress object
    for ert in source:
        style = pd.create_line_style()
        for g in groups:
            g.add_line_style(ert, style)
        x = ert.ert[:, 0]
        style["x"] = x
        x_axis.register_array(x)
        y = ert.ert[:, 1]
        y_axis.register_array(y)
        style["y"] = y
        plot_list.append(style)
    del source

    font_size_0: Final[float] = importance_to_font_size_func(0)

    # set up the graphics area
    axes: Final[Axes] = pu.get_axes(figure)
    axes.tick_params(axis="x", labelsize=font_size_0)
    axes.tick_params(axis="y", labelsize=font_size_0)

    # draw the grid
    if x_grid or y_grid:
        grid_lwd = importance_to_line_width_func(-1)
        if x_grid:
            axes.grid(axis="x", color=pd.GRID_COLOR, linewidth=grid_lwd)
        if y_grid:
            axes.grid(axis="y", color=pd.GRID_COLOR, linewidth=grid_lwd)

    max_y = y_axis.get_pinf_replacement()

    # plot the lines
    for line in plot_list:
        y = line["y"]
        if np.isposinf(y[0]):
            y = y.copy()
            y[0] = max_y
            line["y"] = y
        axes.step(where="post", **line)
    del plot_list

    x_axis.apply(axes, "x")
    y_axis.apply(axes, "y")

    if legend:
        handles: list[Artist] = []

        for g in groups:
            g.add_to_legend(handles.append)
            g.has_style = False

        if instances.has_style:
            instances.add_to_legend(handles.append)
        if algorithms.has_style:
            algorithms.add_to_legend(handles.append)

        axes.legend(loc="upper right",
                    handles=handles,
                    labelcolor=[art.color if hasattr(art, "color")
                                else pd.COLOR_BLACK for art in handles],
                    fontsize=font_size_0)

    pu.label_axes(axes=axes,
                  x_label=x_label(x_dim) if callable(x_label) else x_label,
                  x_label_inside=x_label_inside,
                  x_label_location=1,
                  y_label=y_label(y_dim) if callable(y_label) else y_label,
                  y_label_inside=y_label_inside,
                  y_label_location=0,
                  font_size=font_size_0)
    return axes



"""A unary operator flipping each bit with probability m/n."""
from typing import Callable, Final

import numpy as np
from numpy.random import Generator

from moptipy.api.operators import Op1
from moptipy.utils.nputils import int_range_to_dtype
from moptipy.utils.types import type_error


class Op1MoverNflip(Op1):
    """
    This unary search operation flips each bit with probability of `m/n`.

    For bit strings of length `n`, draw the number `z` of bits to flip from a
    binomial distribution with `p=m/n`. If `at_least_1` is set to `True`, then
    we repeat drawing `z` until `z>0`.
    """

    def __init__(self, n: int, m: int = 1, at_least_1: bool = True):
        """
        Initialize the operator.

        :param n: the length of the bit strings
        :param m: the factor for computing the probability of flipping
            the bits
        :param at_least_1: should at least one bit be flipped?
        """
        super().__init__()
        if not isinstance(n, int):
            raise type_error(n, "n", int)
        if not isinstance(m, int):
            raise type_error(m, "m", int)
        if not (0 < m <= n):
            raise ValueError(f"m must be in 1..{n}, but is {m}.")
        #: the value of m in p=m/n
        self.__m: Final[int] = m
        if not isinstance(at_least_1, bool):
            raise type_error(at_least_1, "at_least_1", bool)
        #: is it OK to not flip any bit?
        self.__none_is_ok: Final[bool] = not at_least_1
        #: the internal permutation
        self.__permutation: Final[np.ndarray] = np.array(
            range(n), dtype=int_range_to_dtype(0, n - 1))

    def op1(self, random: Generator, dest: np.ndarray, x: np.ndarray) -> None:
        """
        Copy `x` into `dest` and flip each bit with probability m/n.

        This method will first copy `x` to `dest`. Then it will flip each bit
        in `dest` with probability `m/n`, where `n` is the length of `dest`.
        Regardless of the probability, at least one bit will always be
        flipped if self.at_least_1 is True.

        :param self: the self pointer
        :param random: the random number generator
        :param dest: the array to be shuffled
        :param x: the existing point in the search space
        """
        np.copyto(dest, x)  # copy source to destination
        length: Final[int] = len(dest)  # get n
        p: Final[float] = self.__m / length  # probability to flip bit
        none_is_ok: Final[bool] = self.__none_is_ok

        flips: int  # the number of bits to flip
        rbi: Final[Callable[[int, float], int]] = random.binomial
        while True:
            flips = rbi(length, p)  # compute the number of bits to flip
            if flips > 0:
                break  # we will flip some bit
            if none_is_ok:
                return  # we will flip no bit

        permutation: Final[np.ndarray] = self.__permutation
        i: int = length
        end: Final[int] = length - flips
        ri: Final[Callable[[int], int]] = random.integers
        while i > end:  # we iterate from i=length down to end=length-flips
            k = ri(i)  # get index of next bit index in permutation
            i -= 1  # decrease i
            idx = permutation[k]  # get index of bit to flip and move to end
            permutation[i], permutation[k] = idx, permutation[i]
            dest[idx] = not dest[idx]  # flip bit

    def __str__(self) -> str:
        """
        Get the name of this unary operator.

        :return: "m_over_n_flip"
        """
        return f"flipB{self.__m}{'n' if self.__none_is_ok else ''}"



"""
The implementation of the basic hill climbing algorithm `hc`.

The algorithm starts by applying the nullary search operator, an
implementation of :meth:`~moptipy.api.operators.Op0.op0`, to sample one fully
random solution. This is the first best-so-far solution. In each step, it
applies the unary operator, an implementation of
:meth:`~moptipy.api.operators.Op1.op1`, to the best-so-far solution to obtain
a new, similar solution. If this new solution is strictly better than the
current best-so-far solution, it replaces this solution. Otherwise, it is
discarded.

The hill climbing algorithm is a simple local search that only accepts
strictly improving moves. It is thus similar to the randomized local search
(`rls`) implemented in :class:`~moptipy.algorithms.so.rls.RLS`, which, however,
accepts non-deteriorating moves. We also provide `hcr`, a variant of the hill
climber that restarts automatically with a certain number of moves were not
able to improve the current best-so-far solution in class :class:`~moptipy.\
algorithms.so.hill_climber_with_restarts.HillClimberWithRestarts`.

1. Stuart Jonathan Russell and Peter Norvig. *Artificial Intelligence: A
   Modern Approach (AIMA)*. 2nd edition. 2002. Upper Saddle River, NJ, USA:
   Prentice Hall International Inc. ISBN: 0-13-080302-2
2. Steven S. Skiena. *The Algorithm Design Manual.* 2nd edition. 2008.
   London, UK: Springer-Verlag. ISBN: 978-1-84800-069-8.
   http://doi.org/10.1007/978-1-84800-070-4.
3. David Stifler Johnson, Christos H. Papadimitriou, and Mihalis Yannakakis.
   How Easy Is Local Search? Journal of Computer and System Sciences.
   37(1):79-100. August 1988. http://doi.org/10.1016/0022-0000(88)90046-3
   http://www2.karlin.mff.cuni.cz/~krajicek/jpy2.pdf
4. James C. Spall. *Introduction to Stochastic Search and Optimization.*
   April 2003. Estimation, Simulation, and Control -- Wiley-Interscience
   Series in Discrete Mathematics and Optimization, volume 6. Chichester, West
   Sussex, UK: Wiley Interscience. ISBN: 0-471-33052-3.
   http://www.jhuapl.edu/ISSO/
5. Holger H. Hoos and Thomas Stützle. *Stochastic Local Search: Foundations
   and Applications.* 2005. ISBN: 1493303732. In The Morgan Kaufmann Series in
   Artificial Intelligence. Amsterdam, The Netherlands: Elsevier.
6. Thomas Weise. *Optimization Algorithms.* 2021. Hefei, Anhui, China:
   Institute of Applied Optimization (IAO), School of Artificial Intelligence
   and Big Data, Hefei University. http://thomasweise.github.io/oa/
7. Thomas Weise. *Global Optimization Algorithms - Theory and Application.*
   2009. http://www.it-weise.de/projects/book.pdf
"""
from typing import Callable, Final

from numpy.random import Generator

from moptipy.api.algorithm import Algorithm1
from moptipy.api.operators import Op0, Op1
from moptipy.api.process import Process


# start book
class HillClimber(Algorithm1):
    """
    The stochastic hill climbing algorithm only accepts improving moves.

    In each step, a hill climber creates a modified copy `new_x` of the
    current best solution `best_x`. If `new_x` is better than `best_x`,
    it becomes the new `best_x`. Otherwise, it is discarded.
    """

    def solve(self, process: Process) -> None:
        """
        Apply the hill climber to an optimization problem.

        :param process: the black-box process object
        """
        # Create records for old and new point in the search space.
        best_x = process.create()  # record for best-so-far solution
        new_x = process.create()  # record for new solution
        # Obtain the random number generator.
        random: Final[Generator] = process.get_random()

        # Put function references in variables to save time.
        evaluate: Final[Callable] = process.evaluate  # the objective
        op1: Final[Callable] = self.op1.op1  # the unary operator
        should_terminate: Final[Callable] = process.should_terminate

        # Start at a random point in the search space and evaluate it.
        self.op0.op0(random, best_x)  # Create 1 solution randomly and
        best_f: int | float = evaluate(best_x)  # evaluate it.

        while not should_terminate():  # Until we need to quit...
            op1(random, new_x, best_x)  # new_x = neighbor of best_x
            new_f: int | float = evaluate(new_x)
            if new_f < best_f:  # new_x is _better_ than best_x?
                best_f = new_f  # Store its objective value.
                best_x, new_x = new_x, best_x  # Swap best and new.
# end book

    def __init__(self, op0: Op0, op1: Op1) -> None:
        """
        Create the hill climber.

        :param op0: the nullary search operator
        :param op1: the unary search operator
        """
        super().__init__("hc", op0, op1)



"""Test stuff on permutations with repetitions."""

from typing import Any, Callable, Iterable, cast

import numpy as np
from numpy.random import Generator, default_rng

from moptipy.api.operators import Op0, Op1, Op2
from moptipy.spaces.permutations import Permutations
from moptipy.tests.op0 import validate_op0
from moptipy.tests.op1 import validate_op1
from moptipy.tests.op2 import validate_op2


def permutations_for_tests() -> Iterable[Permutations]:
    """
    Get a sequence of permutations with repetitions for tests.

    :returns: the sequence of Permutations
    """
    r = default_rng()
    pwrs: list[Permutations] = [
        Permutations.with_repetitions(2, 2),
        Permutations.with_repetitions(2, 3),
        Permutations.with_repetitions(3, 2),
        Permutations.with_repetitions(3, 3),
        Permutations.with_repetitions(5, 5),
        Permutations.with_repetitions(int(r.integers(6, 10)),
                                      int(r.integers(2, 7))),
        Permutations.with_repetitions(int(r.integers(2, 5)),
                                      int(r.integers(6, 10))),
        Permutations.with_repetitions(int(r.integers(130, 500)),
                                      int(r.integers(2, 200)))]
    r.shuffle(cast(list, pwrs))
    return pwrs


def make_permutation_valid(pwr: Permutations) -> \
        Callable[[Generator, np.ndarray], np.ndarray]:
    """
    Create a function that can make permutations with repetitions valid.

    :param pwr: the permutations
    :returns: the function
    """
    def __make_valid(prnd: Generator,
                     x: np.ndarray,
                     ppp=pwr) -> np.ndarray:
        np.copyto(x, ppp.blueprint)
        prnd.shuffle(x)
        return x

    return __make_valid


def validate_op0_on_1_permutations(
        op0: Op0 | Callable[[Permutations], Op0],
        search_space: Permutations,
        number_of_samples: int | None = None,
        min_unique_samples:
        int | Callable[[int, Permutations], int] | None
        = None) -> None:
    """
    Validate the nullary operator on one `Permutations` instance.

    :param op0: the operator or operator factory
    :param search_space: the search space
    :param number_of_samples: the optional number of samples
    :param min_unique_samples: the optional unique samples
    """
    args: dict[str, Any] = {
        "op0": op0(search_space) if callable(op0) else op0,
        "search_space": search_space,
        "make_search_space_element_valid":
            make_permutation_valid(search_space)
    }
    if number_of_samples is not None:
        args["number_of_samples"] = number_of_samples
    if min_unique_samples is not None:
        args["min_unique_samples"] = min_unique_samples
    validate_op0(**args)


def validate_op0_on_permutations(
        op0: Op0 | Callable[[Permutations], Op0],
        number_of_samples: int | None = None,
        min_unique_samples:
        int | Callable[[int, Permutations], int] | None
        = None) -> None:
    """
    Validate the nullary operator on several `Permutations` instances.

    :param op0: the operator or operator factory
    :param number_of_samples: the optional number of samples
    :param min_unique_samples: the optional unique samples
    """
    for pwr in permutations_for_tests():
        validate_op0_on_1_permutations(op0, pwr, number_of_samples,
                                       min_unique_samples)


def validate_op1_on_1_permutations(
        op1: Op1 | Callable[[Permutations], Op1],
        search_space: Permutations,
        number_of_samples: int | None = None,
        min_unique_samples:
        int | Callable[[int, Permutations], int] | None
        = None) -> None:
    """
    Validate the unary operator on one `Permutations` instance.

    :param op1: the operator or operator factory
    :param search_space: the search space
    :param number_of_samples: the optional number of samples
    :param min_unique_samples: the optional unique samples
    """
    args: dict[str, Any] = {
        "op1": op1(search_space) if callable(op1) else op1,
        "search_space": search_space,
        "make_search_space_element_valid":
            make_permutation_valid(search_space)
    }
    if number_of_samples is not None:
        args["number_of_samples"] = number_of_samples
    if min_unique_samples is not None:
        args["min_unique_samples"] = min_unique_samples
    validate_op1(**args)


def validate_op1_on_permutations(
        op1: Op1 | Callable[[Permutations], Op1],
        number_of_samples: int | None = None,
        min_unique_samples:
        int | Callable[[int, Permutations], int] | None
        = None) -> None:
    """
    Validate the unary operator on several `Permutations` instances.

    :param op1: the operator or operator factory
    :param number_of_samples: the optional number of samples
    :param min_unique_samples: the optional unique samples
    """
    for pwr in permutations_for_tests():
        validate_op1_on_1_permutations(op1, pwr, number_of_samples,
                                       min_unique_samples)


def validate_op2_on_1_permutations(
        op2: Op2 | Callable[[Permutations], Op2],
        search_space: Permutations,
        number_of_samples: int | None = None,
        min_unique_samples:
        int | Callable[[int, Permutations], int] | None
        = None) -> None:
    """
    Validate the binary operator on one `Permutations` instance.

    :param op2: the operator or operator factory
    :param search_space: the search space
    :param number_of_samples: the optional number of samples
    :param min_unique_samples: the optional unique samples
    """
    args: dict[str, Any] = {
        "op2": op2(search_space) if callable(op2) else op2,
        "search_space": search_space,
        "make_search_space_element_valid":
            make_permutation_valid(search_space)
    }
    if number_of_samples is not None:
        args["number_of_samples"] = number_of_samples
    if min_unique_samples is not None:
        args["min_unique_samples"] = min_unique_samples
    validate_op2(**args)


def validate_op2_on_permutations(
        op2: Op2 | Callable[[Permutations], Op2],
        number_of_samples: int | None = None,
        min_unique_samples:
        int | Callable[[int, Permutations], int] | None
        = None) -> None:
    """
    Validate the binary operator on several `Permutations` instances.

    :param op2: the operator or operator factory
    :param number_of_samples: the optional number of samples
    :param min_unique_samples: the optional unique samples
    """
    for pwr in permutations_for_tests():
        validate_op2_on_1_permutations(op2, pwr, number_of_samples,
                                       min_unique_samples)



"""Validate selection algorithms."""

from math import inf
from typing import Final, Iterable

from numpy.random import Generator, default_rng

from moptipy.algorithms.modules.selection import (
    FitnessRecord,
    Selection,
    check_selection,
)
from moptipy.tests.component import validate_component
from moptipy.utils.types import type_error


class _FRecord(FitnessRecord):
    """The internal Fitness-record."""

    def __init__(self, tag: int) -> None:
        """Initialize."""
        #: the fitness
        self.fitness: int | float = inf
        #: the tag
        self.tag: Final[int] = tag

    def __str__(self) -> str:
        """Get the string describing this record."""
        return f"{self.tag}/{self.fitness}"


def __join(sets: Iterable[Iterable[int]],
           lower_limit: int | float = -inf,
           upper_limit: int | float = inf) -> Iterable[int]:
    """
    Joint iterables preserving unique values.

    :param sets: the iterables
    :param lower_limit: the lower limit
    :param upper_limit: the upper limit
    :returns: the joint iterable
    """
    if not isinstance(sets, Iterable):
        raise type_error(sets, "sets", Iterable)
    if not isinstance(lower_limit, (int, float)):
        raise type_error(lower_limit, "lower_limit", (int, float))
    if not isinstance(upper_limit, (int, float)):
        raise type_error(upper_limit, "upper_limit", (int, float))
    if upper_limit < lower_limit:
        raise ValueError(
            f"lower_limit={lower_limit} but upper_limit={upper_limit}")
    x: Final[set[int]] = set()
    for it in sets:
        if not isinstance(it, Iterable):
            raise type_error(it, "it", Iterable)
        x.update([int(i) for i in it if lower_limit <= i <= upper_limit])
    return list(x)


def validate_selection(selection: Selection,
                       without_replacement: bool = False,
                       lower_source_size_limit: int = 0,
                       upper_source_size_limit: int = 999999) -> None:
    """
    Validate a selection algorithm.

    :param selection: the selection algorithm
    :param without_replacement: is this selection algorithm without
        replacement, i.e., can it select each element at most once?
    :param lower_source_size_limit: the lower limit of the source size
    :param upper_source_size_limit: the upper limit for the source size
    """
    check_selection(selection)
    validate_component(selection)

    if not isinstance(without_replacement, bool):
        raise type_error(without_replacement, "without_replacement", bool)
    if not isinstance(lower_source_size_limit, int):
        raise type_error(lower_source_size_limit,
                         "lower_source_size_limit", int)
    if not 0 <= lower_source_size_limit < 1000:
        raise ValueError(f"lower_source_size_limit={lower_source_size_limit}")
    if not isinstance(upper_source_size_limit, int):
        raise type_error(upper_source_size_limit,
                         "upper_source_size_limit", int)
    if upper_source_size_limit < lower_source_size_limit:
        raise ValueError(
            f"upper_source_size_limit={upper_source_size_limit} while "
            f"lower_source_size_limit={lower_source_size_limit}")

    random: Final[Generator] = default_rng()
    source: Final[list[FitnessRecord]] = []
    copy: Final[dict[int, list]] = {}
    dest: Final[list[FitnessRecord]] = []
    tag: int = 0

    for source_size in __join([range(1, 10), [16, 32, 50, 101],
                               random.choice(100, 4, False),
                               [lower_source_size_limit]],
                              lower_limit=max(1, lower_source_size_limit),
                              upper_limit=upper_source_size_limit):
        for dest_size in __join([
            [1, 2, 3, source_size, 2 * source_size],
            random.choice(min(6, 4 * source_size),
                          min(6, 4 * source_size), False)],
                upper_limit=source_size if without_replacement else inf,
                lower_limit=1):

            source.clear()
            copy.clear()
            dest.clear()

            # choose the fitness function
            fit_choice = random.integers(3)
            if fit_choice == 0:
                def fitness(  # noqa
                        value=int(random.integers(-10, 10))):  # type: ignore
                    return value
            elif fit_choice == 1:
                def fitness(  # noqa
                        limit=random.integers(100) + 1):  # type: ignore
                    return int(random.integers(limit))
            else:
                def fitness():  # type: ignore # noqa
                    return float(random.uniform())

            for _ in range(source_size):
                tag += 1
                r1 = _FRecord(tag)
                r1.fitness = fitness()
                r2 = _FRecord(tag)
                r2.fitness = r1.fitness
                source.append(r1)
                copy[r2.tag] = [r2, 0, False]

            # perform the selection
            selection.select(source, dest.append, dest_size, random)

            if len(dest) != dest_size:
                raise ValueError(
                    f"expected {selection} to select {dest_size} elements "
                    f"from {source_size}, but got {len(dest)} instead")
            if len(source) != source_size:
                raise ValueError(
                    f"selection {selection} changed length {source_size} "
                    f"of source list to {len(source)}")
            for ele in dest:
                if not isinstance(ele, _FRecord):
                    raise type_error(ele, "element in dest", _FRecord)
                if ele.tag not in copy:
                    raise ValueError(
                        f"element with tag {ele.tag} does not exist in "
                        f"source but was selected by {selection}?")
                check = copy[ele.tag]
                if check[0].fitness != ele.fitness:
                    raise ValueError(
                        f"fitness of source element {check[0].fitness} has "
                        f"been changed to {ele.fitness} by {selection} "
                        "in dest")
                check[1] += 1
                if without_replacement and (check[1] > 1):
                    raise ValueError(
                        f"{selection} is without replacement, but selected "
                        f"element with tag {ele.tag} at least twice!")
            for ele in source:
                if not isinstance(ele, _FRecord):
                    raise type_error(ele, "element in source", _FRecord)
                if ele.tag not in copy:
                    raise ValueError(
                        f"element with tag {ele.tag} does not exist in "
                        f"source but was created by {selection}?")
                check = copy.get(ele.tag)
                if check[0].fitness != ele.fitness:
                    raise ValueError(
                        f"fitness of source element {check[0].fitness} has "
                        f"been changed to {ele.fitness} by {selection} "
                        "in source")
                if check[2]:
                    raise ValueError(
                        f"element with tag {ele.tag} has been replicated "
                        f"by {selection} in source?")
                check[2] = True
            for check in copy.values():
                if not check[2]:
                    raise ValueError(
                        f"element with tag {check[0].tag} somehow lost?")



"""A nullary operator filling a vector with uniformly distribute values."""

from typing import Final

import numpy as np
from numpy.random import Generator

from moptipy.api.operators import Op0
from moptipy.spaces.vectorspace import VectorSpace
from moptipy.utils.logger import KeyValueLogSection
from moptipy.utils.types import type_error


class Op0Uniform(Op0):
    """Fill a vector with uniformly distributed random values."""

    def __init__(self, space: VectorSpace) -> None:
        """
        Initialize the operator.

        :param space: the vector space
        """
        super().__init__()
        if not isinstance(space, VectorSpace):
            raise type_error(space, "space", VectorSpace)
        #: store the space
        self.space: Final[VectorSpace] = space

    def __str__(self) -> str:
        """
        Get the name of this operator.

        :returns: the name of this space
        """
        return "uniform"

    def op0(self, random: Generator, dest: np.ndarray) -> None:
        """
        Fill the string `dest` with random values.

        :param random: the random number generator
        :param dest: the vector to be filled. Afterwards it contains uniformly
            distributed random values
        """
        sp: Final[VectorSpace] = self.space
        np.copyto(dest, random.uniform(
            sp.lower_bound, sp.upper_bound, sp.dimension))

    def log_parameters_to(self, logger: KeyValueLogSection) -> None:
        """
        Log the parameters of this operator to the given logger.

        :param logger: the logger for the parameters
        """
        super().log_parameters_to(logger)
        self.space.log_bounds(logger)



"""Here we implement a space implementation for `Gantt` charts."""
from math import factorial
from typing import Final

import numpy as np

from moptipy.api.space import Space
from moptipy.examples.jssp.gantt import Gantt
from moptipy.examples.jssp.instance import SCOPE_INSTANCE, Instance
from moptipy.utils.logger import CSV_SEPARATOR, KeyValueLogSection
from moptipy.utils.nputils import (
    KEY_NUMPY_TYPE,
    int_range_to_dtype,
    val_numpy_type,
)
from moptipy.utils.types import type_error

#: the array shape
KEY_SHAPE: Final[str] = "shape"


def gantt_space_size(jobs: int, machines: int) -> int:
    """
    Compute the size of the Gantt space.

    :param jobs: the number of jobs
    :param machines: the number of machines
    :return: the size of the search

    >>> print(gantt_space_size(8, 5))
    106562062388507443200000
    """
    if not isinstance(jobs, int):
        raise type_error(jobs, "number of jobs", int)
    if jobs <= 0:
        raise ValueError(f"Number of jobs must be > 0, but is {jobs}.")
    if not isinstance(machines, int):
        raise type_error(machines, "number of machines", int)
    if machines <= 0:
        raise ValueError(
            f"Number of machines must be > 0, but is {machines}.")
    return factorial(jobs) ** machines


# start book
class GanttSpace(Space):
    """An implementation of the `Space` API of for `Gantt` charts."""

    def __init__(self, instance: Instance) -> None:
        # end book
        """
        Create a Gantt chart space.

        :param instance: the JSSP instance
        """
        if not isinstance(instance, Instance):
            ValueError("Must provide valid JSSP instance, "
                       f"but passed in a {type(instance)}.")
        #: The JSSP Instance to which the Gantt record apply.
        self.instance: Final[Instance] = instance  # +book
        #: The shape for the Gantt chart arrays.
        self.shape: Final[tuple[int, int, int]] = (  # +book
            instance.machines, instance.jobs, 3)  # +book
        #: the data to be used for Gantt charts
        self.dtype: Final[np.dtype] = int_range_to_dtype(
            min_value=0, max_value=instance.makespan_upper_bound)
        #: fast call function forward
        self.copy = np.copyto  # type: ignore # +book

    def create(self) -> Gantt:  # +book
        """
        Create a Gantt chart object without assigning jobs to machines.

        :return: the Gantt chart
        """
        return Gantt(self)  # +book

    def to_str(self, x: Gantt) -> str:  # +book
        """
        Convert a Gantt chart to a string.

        :param x: the Gantt chart
        :return: a string corresponding to the flattened `Gantt` array
        """
        return CSV_SEPARATOR.join([str(xx) for xx in np.nditer(x)])  # +book

    def is_equal(self, x1: Gantt, x2: Gantt) -> bool:  # +book
        """
        Check if two Gantt charts have the same contents.

        :param x1: the first chart
        :param x2: the second chart
        :return: `True` if both charts are for the same instance and have the
            same structure
        """
        # start book
        return (x1.instance is x2.instance) and np.array_equal(x1, x2)
        # end book

    def from_str(self, text: str) -> Gantt:  # +book
        """
        Convert a string to a Gantt chart.

        :param text: the string
        :return: the Gantt chart
        """
        if not isinstance(text, str):
            raise type_error(text, "instance text", str)
        # start book
        x: Final[Gantt] = self.create()
        np.copyto(x, np.fromstring(
            text, dtype=self.dtype, sep=CSV_SEPARATOR)
            .reshape(self.shape))
        self.validate(x)  # -book
        return x
        # end book

    def validate(self, x: Gantt) -> None:  # +book
        """
        Check if a Gantt chart if valid and feasible.

        This means that the operations of the jobs must appear in the right
        sequences and must not intersect in any way.
        The only exception are operations that need 0 time units. They are
        permitted to appear wherever.

        :param x: the Gantt chart
        :raises TypeError: if any component of the chart is of the wrong type
        :raises ValueError: if the Gantt chart is not feasible or the makespan
            is wrong
        """
        # start book
        # Checks if a Gantt chart if valid and feasible.
        if not isinstance(x, Gantt):
            raise type_error(x, "x", Gantt)
        # the rest of the checks is not printed for brevity reasons...
        # end book
        inst: Final[Instance] = self.instance
        if inst is not x.instance:
            raise ValueError(
                f"x.instance must be {inst} but is {x.instance}.")
        jobs: Final[int] = inst.jobs
        machines: Final[int] = inst.machines
        if len(x.shape) != 3:
            raise ValueError("x must be three-dimensional, "
                             f"but is {len(x.shape)}-dimensional.")
        if x.shape[0] != machines:
            raise ValueError(
                f"x must have {machines} rows for instance {inst.name}, "
                f"but has {x.shape[0]}.")

        if x.shape[1] != jobs:
            raise ValueError(
                f"x must have {jobs} columns for instance {inst.name}, "
                f"but has {x.shape[1]}.")
        if x.shape[2] != 3:
            raise ValueError("x must have three values per cell, "
                             f"but has {x.shape[2]}.")
        if x.dtype != self.dtype:
            raise ValueError(
                f"x.dtype should be {self.dtype} for instance "
                f"{inst.name} but is {x.dtype}.")

        # now check the sequence on operations per machine
        # we check for overlaps and incorrect operation times
        jobs_done: Final[set[int]] = set()
        for machinei in range(machines):
            jobs_done.clear()
            last_end = 0
            last_name = "[start]"
            for jobi in range(jobs):
                job = int(x[machinei, jobi, 0])
                start = int(x[machinei, jobi, 1])
                end = int(x[machinei, jobi, 2])
                if not (0 <= job < jobs):
                    raise ValueError(
                        f"job {job} invalid for machine {machinei} on "
                        f"instance {inst.name} with {jobs} jobs: "
                        f"{x[machinei]}")
                if job in jobs_done:
                    raise ValueError(
                        f"job {job} appears twice on machine {machinei} "
                        f"for instance {inst.name}: {x[machinei]}")
                jobs_done.add(job)
                time = -1
                for z in range(machines):
                    if inst[job, z, 0] == machinei:
                        time = int(inst[job, z, 1])
                        break
                if time < 0:
                    raise ValueError(
                        f"Did not find machine {machinei} for job {job} in "
                        f"instance {inst.name}: {x[machinei]}, "
                        f"{inst[machinei]}.")
                if (end - start) != time:
                    raise ValueError(
                        f"job {job} should need {time} time units on machine "
                        f"{machinei} for instance {inst.name}, but starts at "
                        f"{start} and ends at {end}: {x[machinei]}, "
                        f"{inst[job]}.")
                if time <= 0:
                    continue  # job requires zero time, can skip
                if start < last_end:
                    raise ValueError(
                        f"job {job} starts at {start} on machine {machinei}, "
                        f"for instance {inst.name} but cannot before "
                        f"{last_name}: {x[machinei]}.")
                last_end = end

        # now check the single jobs
        # we again check for operation overlaps and incorrect timing
        for jobi in range(jobs):
            done = [(start, end, machine)
                    for machine in range(machines)
                    for (idx, start, end) in x[machine, :, :] if idx == jobi]
            done.sort()
            if len(done) != machines:
                raise ValueError(
                    f"Job {jobi} appears only {len(done)} times instead of "
                    f"{machines} times on instance {inst.name}.")

            # we allow operations of length 0 to appear at any position
            last_end = 0
            last_machine = "[start]"
            done_i = 0
            machine_i = 0
            while True:
                if machine_i < machines:
                    machine, time = inst[jobi, machine_i]
                else:
                    machine = time = -1
                if done_i < machines:
                    start = int(done[machine_i][0])
                    end = int(done[machine_i][1])
                    used_machine = int(done[machine_i][2])
                else:
                    start = end = used_machine = -1

                needed = end - start
                if needed < 0:
                    raise ValueError(
                        f"Operation {machine_i} of job {jobi} scheduled "
                        f"from {start} to {end}?")
                if machine != used_machine:
                    # This is only possible for operations that require zero
                    # time units. We will skip such operations in the checks.
                    if (time == 0) and (machine != -1):
                        machine_i += 1
                        continue
                    if (needed == 0) and (machine != -1):
                        done_i += 1
                        continue
                    raise ValueError(
                        f"Machine at index {done_i} of job {jobi} "
                        f"must be {machine} for instance {inst.name}, "
                        f"for {time} time units, but is {used_machine}"
                        f"from {start} to {end}.")
                if machine == -1:
                    if (machine_i < machines) or (done_i < machines):
                        raise ValueError(  # this should never be possible
                            f"Done {machine_i + 1} machines for job {jobi}, "
                            f"which has {done_i + 1} operations done??")
                    break  # we can stop

                # ok, we are at a regular operation
                if needed != time:
                    raise ValueError(
                        f"Job {jobi} must be processed on {machine} for "
                        f"{time} time units on instance {inst.name}, but "
                        f"only {needed} are used from {start} to {end}.")
                if needed < 0:
                    raise ValueError(
                        f"Processing time of job {jobi} on machine {machine} "
                        f"cannot be < 0 for instance {inst.name}, but "
                        f"is {needed}.")

                if start < last_end:
                    raise ValueError(
                        f"Processing time window [{start},{end}] on "
                        f"machine {machine} for job {jobi} on instance "
                        f"{inst.name} intersects with last operation end"
                        f"{last_end} on machine {last_machine}.")

                last_end = end
                last_machine = machine
                machine_i += 1
                done_i += 1

        maxtime: Final[int] = int(x[:, -1, 2].max())
        if maxtime < inst.makespan_lower_bound:
            raise ValueError(
                f"Makespan {maxtime} computed, which is smaller than the "
                f"lower bound {inst.makespan_lower_bound} of the JSSP "
                f"instance '{inst}'.")
        if maxtime > inst.makespan_upper_bound:
            raise ValueError(
                f"Makespan {maxtime} computed, which is larger than the "
                f"upper bound {inst.makespan_upper_bound} of the JSSP "
                f"instance '{inst}'.")

    def n_points(self) -> int:
        """
        Get the number of possible Gantt charts without useless delays.

        :return: `factorial(jobs) ** machines`

        >>> print(GanttSpace(Instance.from_resource("demo")).n_points())
        7962624
        """
        return gantt_space_size(self.instance.jobs, self.instance.machines)

    def __str__(self) -> str:
        """
        Get the name of the Gantt space.

        :return: the name

        >>> space = GanttSpace(Instance.from_resource("abz7"))
        >>> print(space)
        gantt_abz7
        """
        return f"gantt_{self.instance}"

    def log_parameters_to(self, logger: KeyValueLogSection) -> None:
        """
        Log the parameters of the Gantt space to the given logger.

        :param logger: the logger for the parameters
        """
        super().log_parameters_to(logger)
        logger.key_value(KEY_SHAPE, repr(self.shape))
        logger.key_value(KEY_NUMPY_TYPE, val_numpy_type(self.dtype))
        with logger.scope(SCOPE_INSTANCE) as kv:
            self.instance.log_parameters_to(kv)
